import numpy as np
from .fitmulticomponent import fobj_elv, fobj_ell, fobj_hazb
from scipy.optimize import minimize, minimize_scalar
from ..saft import saftvrmie


def fobj_kij(kij, mix, datavle=None, datalle=None, datavlle=None):

    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)
    eos = saftvrmie(mix)

    error = 0.
    if datavle is not None:
        error += fobj_elv(eos, *datavle)
    if datalle is not None:
        error += fobj_ell(eos, *datalle)
    if datavlle is not None:
        error += fobj_hazb(eos, *datavlle)
    return error


def fit_kij(kij_bounds, mix, datavle=None, datalle=None, datavlle=None,
            minimize_options={}):
    """
    fit_kij attemps to fit kij to VLE, LLE, VLLE

    Parameters
    ----------
    kij_bounds: tuple
        bounds for kij correction
    mix: object
        binary mixture
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
     minimize_options: dict
        Dictionary of any additional spefication for scipy minimize_scalar

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize_scalar(fobj_kij, kij_bounds, args=(mix, datavle, datalle,
                          datavlle), **minimize_options)
    return fit


def fobj_asso(x, mix, datavle=None, datalle=None, datavlle=None):
    kij, lij = x
    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)
    Lij = np.array([[0, lij], [lij, 0]])
    mix.lij_saft(Lij)
    eos = saftvrmie(mix)

    error = 0.
    if datavle is not None:
        error += fobj_elv(eos, *datavle)
    if datalle is not None:
        error += fobj_ell(eos, *datalle)
    if datavlle is not None:
        error += fobj_hazb(eos, *datavlle)
    return error


def fit_asso(x0, mix, datavle=None, datalle=None, datavlle=None,
             minimize_options={}):
    """
    fit_asso attemps to fit kij and lij to VLE, LLE, VLLE

    Parameters
    ----------
    x0: array
        initial values for kij and lij
    mix: object
        binary mixture
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
     minimize_options: dict
        Dictionary of any additional spefication for scipy minimize

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize(fobj_asso, x0, args=(mix, datavle, datalle, datavlle),
                   **minimize_options)
    return fit


def fobj_cross(x, mix, assoc, datavle=None, datalle=None, datavlle=None):
    kij, rc = x
    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)

    eos = saftvrmie(mix)
    eos.eABij[0, 1] = mix.eAB[assoc] / 2
    eos.eABij[1, 0] = mix.eAB[assoc] / 2
    eos.rcij[0, 1] = rc * 1e-10
    eos.rcij[1, 0] = rc * 1e-10

    error = 0.
    if datavle is not None:
        error += fobj_elv(eos, *datavle)
    if datalle is not None:
        error += fobj_ell(eos, *datalle)
    if datavlle is not None:
        error += fobj_hazb(eos, *datavlle)
    return error


def fit_cross(x0, mix, assoc, datavle=None, datalle=None, datavlle=None,
              minimize_options={}):
    """
    fit_cross attemps to fit kij and rcij to VLE, LLE, VLLE

    Parameters
    ----------
    x0 : array
        initial values for kij and rc
    mix: object
        binary mixture
    assoc : int
        index of associating component
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
    minimize_options: dict
        Dictionary of any additional spefication for scipy minimize

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize(fobj_cross, x0, args=(mix, assoc, datavle, datalle,
                   datavlle), **minimize_options)
    return fit



from __future__ import division, print_function, absolute_import
from .mixtures.saftvrmiemix import saftvrmie_mix
from .pure.saftvrmie import saftvrmie_pure


def saftvrmie(mix_or_component):
    '''
    Returns SAFT-VR-Mie EoS object.

    Parameters
    ----------
    mix_or_component : object
        :class:`SGTPy.mixture` or :class:`SGTPy.component` object
    Returns
    -------
    eos : object
        SAFT-VR-Mie EoS object
    '''
    nc = mix_or_component.nc
    if nc == 1:
        eos = saftvrmie_pure(mix_or_component)
    else:
        eos = saftvrmie_mix(mix_or_component)
    return eos



import numpy as np
from .gmie_aux import gdHS, dgdHS_drho, d2gdHS_drho
from .gmie_aux import g1sigma, dg1sigma_drho, d2g1sigma_drho
from .gmie_aux import g2sigma, dg2sigma_drho, d2g2sigma_drho


def gmie(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
         x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho, beta, eps,
         c, ms):

    be = beta * eps

    ghs = gdHS(x0, eta)
    g1 = g1sigma(x0_a1, a1sb_a1, da1m_deta,  eta, d, deta_drho, rho, eps,
                 c, ms)
    g2 = g2sigma(x03, nsigma, alpha, tetha, x0_a2, a1sb_a2, da2m_new_deta,
                 khs, eta,  d, deta_drho, rho,  eps, c, ms)

    gm = ghs*np.exp((be*g1+be**2*g2)/ghs)
    return gm


def lngmie(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
           x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho, beta, eps,
           c, ms):
    be = beta * eps
    ghs = gdHS(x0, eta)
    g1 = g1sigma(x0_a1, a1sb_a1, da1m_deta,  eta, d, deta_drho, rho, eps,
                 c, ms)
    g2 = g2sigma(x03, nsigma, alpha, tetha, x0_a2, a1sb_a2, da2m_new_deta,
                 khs, eta,  d, deta_drho, rho, eps, c, ms)

    lng = np.log(ghs) + (be*g1+be**2*g2)/ghs
    return lng


def dlngmie_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
                 x0_a2, a1sb_a2, da2m_new_deta, khs, d, drho, rho, beta,
                 eps, c, ms):
    be = beta * eps

    ghs, dghs = dgdHS_drho(x0, eta, drho)
    g1, dg1 = dg1sigma_drho(x0_a1, a1sb_a1, da1m_deta,  eta, d, drho, rho,
                            eps, c, ms)
    g2, dg2 = dg2sigma_drho(x03, nsigma, alpha, tetha, x0_a2, a1sb_a2,
                            da2m_new_deta, khs, eta, d, drho, rho,  eps, c, ms)

    lng = np.log(ghs) + (be*g1+be**2*g2)/ghs

    dlng = be * ghs * (dg1 + be * dg2)
    dlng += dghs * (ghs - be * (g1 + be * g2))
    dlng /= ghs**2

    return np.hstack([lng, dlng])


def d2lngmie_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha,
                  tetha, x0_a2, a1sb_a2, da2m_new_deta, khs, d, drho, rho,
                  beta, eps, c, ms):

    be = beta * eps

    ghs, dghs, d2ghs = d2gdHS_drho(x0, eta, drho)
    g1, dg1, d2g1 = d2g1sigma_drho(x0_a1, a1sb_a1, da1m_deta, eta, d, drho,
                                   rho, eps, c, ms)
    g2, dg2, d2g2 = d2g2sigma_drho(x03, nsigma, alpha, tetha, x0_a2, a1sb_a2,
                                   da2m_new_deta, khs, eta, d, drho, rho,
                                   eps, c, ms)

    lng = np.log(ghs) + (be*g1+be**2*g2)/ghs

    dlng = be * ghs * (dg1 + be * dg2)
    dlng += dghs * (ghs - be * (g1 + be * g2))
    dlng /= ghs**2

    d2lng = 2*be * dghs**2 * (g1 + be * g2)
    d2lng += ghs**2*(d2ghs + be * (d2g1 + be * d2g2))
    d2lng += ghs*(-dghs*(2*be*(dg1 + be*dg2) + dghs)-be*(g1 + be*g2)*d2ghs)
    d2lng /= ghs**3

    return np.hstack([lng, dlng, d2lng])



import numpy as np


# Equation A17
def xhi_eff(xhix, cictes):
    xhix_vec = np.array([xhix, xhix**2, xhix**3, xhix**4])
    xhieff = np.matmul(cictes, xhix_vec)
    return xhieff


def dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00):
    xhix_vec = np.array([1., 2 * xhix, 3*xhix**2, 4*xhix**3])
    dxhieff = dxhix_dxhi00 * np.matmul(cictes, xhix_vec)
    return dxhieff


def d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00):
    xhix_vec = np.array([0, 2, 6*xhix, 12*xhix**2])
    dxhieff = (dxhix_dxhi00)**2 * np.matmul(cictes, xhix_vec)
    return dxhieff


def d3xhieff_dxhi00(xhix, cictes, dxhix_dxhi00):
    xhix_vec = np.array([0, 0., 6., 24*xhix])
    dxhieff = (dxhix_dxhi00)**3 * np.matmul(cictes, xhix_vec)
    return dxhieff


def dxhieff_dx(xhix, cictes, dxhix_dx):
    xhix_vec = np.array([1., 2 * xhix, 3*xhix**2, 4*xhix**3])
    dxhieff = np.multiply.outer(dxhix_dx, np.matmul(cictes, xhix_vec))
    return dxhieff


def dxhieff_dx_dxhi00(xhix, cictes,  dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):
    xhix_vec = np.array([1., 2 * xhix, 3*xhix**2, 4*xhix**3])
    aux1 = np.matmul(cictes, xhix_vec)

    ter1 = np.multiply.outer(dxhix_dx_dxhi00, aux1)

    xhix_vec2 = np.array([0., 2, 6*xhix, 12*xhix**2])
    ter2 = np.multiply.outer(dxhix_dx, np.matmul(cictes, xhix_vec2))*dxhix_dxhi00
    return ter2 + ter1


# Equation A16
def J_lam(x0, lam):
    lam3 = 3. - lam
    lam4 = 4. - lam
    J = (lam3*x0**lam4 - lam4*x0**lam3 + 1.)/(lam3 * lam4)
    return J


# Equation A14
def I_lam(x0, lam):
    lam3 = 3. - lam
    I = (x0**lam3-1.) / lam3
    return I


# Equation A21
def kHS(xhix):
    k = (1-xhix)**4
    k /= 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4
    return k


def dkHS_dxhi00(xhix, dxhix_dxhi00):
    den = 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4

    k = (1-xhix)**4
    k /= den

    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dxhi00
    dk /= den**2
    return k, dk


def d2kHS_dxhi00(xhix, dxhix_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix2**2
    xhix5 = xhix3 * xhix2
    xhix6 = xhix5 * xhix

    den = 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4

    k = (1-xhix)**4
    k /= den

    dk = -2 - 5*xhix + xhix2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dxhi00
    dk /= den**2

    d2k = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    d2k += 77 * xhix4 - 30 * xhix5 + 3*xhix6
    d2k *= 4 * (- 1 + xhix)**2 * dxhix_dxhi00**2
    d2k /= den**3
    return k, dk, d2k


def d3kHS_dxhi00(xhix, dxhix_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix2**2
    xhix5 = xhix3 * xhix2
    xhix6 = xhix5 * xhix

    den = 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4

    k = (1-xhix)**4
    k /= den

    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dxhi00
    dk /= den**2

    d2k = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    d2k += 77 * xhix4 - 30 * xhix5 + 3*xhix6
    d2k *= 4 * (- 1 + xhix)**2 * dxhix_dxhi00**2
    d2k /= den**3

    d3k = -97 - 109 * xhix + 238 * xhix2 - 352 * xhix4
    d3k += 372 * xhix5 - 210*xhix6 + 77*xhix**7 - 15*xhix**8 + xhix**9
    d3k *= xhix
    d3k -= 13.
    d3k *= 12.
    d3k *= -4 * (- 1 + xhix) * dxhix_dxhi00**3
    d3k /= den**4

    return k, dk, d2k, d3k


def dkHS_dx(xhix, dxhix_dx):
    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dx
    dk /= (1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4)**2
    return dk


def dkHS_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    den = 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4

    k = (1-xhix)**4
    k /= den

    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dxhi00
    dk /= den**2

    dkx = -2 - 5*xhix + xhix**2
    dkx *= -4 * (- 1 + xhix)**3 * dxhix_dx
    dkx /= (1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4)**2

    aux1 = 17 + 82 * xhix + 39 * xhix**2 - 80 * xhix**3
    aux1 += 77 * xhix**4 - 30 * xhix**5 + 3*xhix**6

    aux2 = -2 - 13 * xhix - 27*xhix**2 - 8 * xhix**3
    aux2 += 22 * xhix**4 - 9 * xhix**5 + xhix**6

    dkxxhi = dxhix_dxhi00 * dxhix_dx * aux1
    dkxxhi += -(-1+xhix)*aux2*dxhix_dx_dxhi00

    dkxxhi *= 4 * (- 1 + xhix)**2
    dkxxhi /= den**3
    return k, dk, dkx, dkxxhi


def dkHS_dx_dxhi002(xhix, dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2 * xhix
    xhix4 = xhix2**2
    xhix5 = xhix3 * xhix2
    xhix6 = xhix5 * xhix

    den = 1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4

    k = (1-xhix)**4
    k /= den

    dk = -2 - 5*xhix + xhix**2
    dk *= -4 * (- 1 + xhix)**3 * dxhix_dxhi00
    dk /= den**2

    d2k = 17 + 82 * xhix + 39 * xhix2 - 80 * xhix3
    d2k += 77 * xhix4 - 30 * xhix5 + 3*xhix6
    d2k *= 4 * (- 1 + xhix)**2 * dxhix_dxhi00**2
    d2k /= den**3

    dkx = -2 - 5*xhix + xhix**2
    dkx *= -4 * (- 1 + xhix)**3 * dxhix_dx
    dkx /= (1 + 4*xhix + 4*xhix**2 - 4*xhix**3 + xhix**4)**2

    aux1 = 17 + 82 * xhix + 39 * xhix**2 - 80 * xhix**3
    aux1 += 77 * xhix**4 - 30 * xhix**5 + 3*xhix**6

    aux2 = -2 - 13 * xhix - 27*xhix**2 - 8 * xhix**3
    aux2 += 22 * xhix**4 - 9 * xhix**5 + xhix**6

    dkxxhi = dxhix_dxhi00 * dxhix_dx * aux1
    dkxxhi += -(-1+xhix)*aux2*dxhix_dx_dxhi00

    dkxxhi *= 4 * (- 1 + xhix)**2
    dkxxhi /= den**3
    return k, dk, d2k, dkx, dkxxhi


# Equation A22
def Xi(xhixm, f1, f2, f3):
    x = f1*xhixm + f2*xhixm**5 + f3*xhixm**8
    return x


# Equation A22
def dXi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3):
    xhixm2 = xhixm**2
    xhixm4 = xhixm2**2
    xhixm5 = xhixm4*xhixm
    xhixm7 = xhixm5*xhixm2
    xhixm8 = xhixm4**2

    x = f1*xhixm + f2*xhixm5 + f3*xhixm8

    dx = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dx *= dxhim_dxhi00
    return x, dx


def d2Xi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3):
    xhixm2 = xhixm**2
    xhixm3 = xhixm2*xhixm
    xhixm4 = xhixm2**2
    xhixm5 = xhixm4*xhixm
    xhixm6 = xhixm5*xhixm
    xhixm7 = xhixm5*xhixm2
    xhixm8 = xhixm4**2

    x = f1*xhixm + f2*xhixm5 + f3*xhixm8

    dx = f1 + 5*f2*xhixm4 + 8*f3*xhixm7
    dx *= dxhim_dxhi00

    d2x = 20*f2*xhixm3 + 56*f3*xhixm6
    d2x *= dxhim_dxhi00**2
    return x, dx, d2x


def d3Xi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3):
    x = 60*f2*xhixm**2 + 56*6*f3*xhixm**5
    x *= dxhim_dxhi00**3
    return x


# Equation A22
def dXi_dx(xhixm, dxhim_dx, f1, f2, f3):
    x = f1 + 5*f2*xhixm**4 + 8*f3*xhixm**7
    x = np.multiply.outer(dxhim_dx, x)
    return x



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from .equilibriumresult import EquilibriumResult


def haz_objb(inc, T_P, tipo, model, v0):

    X, W, Y, P_T = np.array_split(inc, 4)

    if tipo == 'T':
        P = P_T
        T = T_P
    elif tipo == 'P':
        T = P_T
        P = T_P

    global vx, vw, vy
    vx, vw, vy = v0

    fugX, vx = model.logfugef(X, T, P, 'L', vx)
    fugW, vw = model.logfugef(W, T, P, 'L', vw)
    fugY, vy = model.logfugef(Y, T, P, 'V', vy)

    K1 = np.exp(fugX-fugY)
    K2 = np.exp(fugX-fugW)
    return np.hstack([K1*X-Y, K2*X-W, X.sum()-1, Y.sum()-1, W.sum()-1])


def vlleb(X0, W0, Y0, P_T, T_P, spec, model, v0=[None, None, None],
          full_output=False):
    '''
    Solves liquid liquid vapour equilibrium for binary mixtures.
    (T,P) -> (x,w,y)

    Parameters
    ----------

    X0 : array_like
        guess composition of phase 1
    W0 : array_like
        guess composition of phase 1
    Y0 : array_like
        guess composition of phase 2
    P_T : float
        absolute temperature [K] or pressure [Pa]
    T_P : floar
        absolute temperature [K] or pressure [Pa]
    spec: string
        'T' if T_P is temperature or 'P' if pressure.
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------

    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector
    var: float
        temperature or pressure, depending of specification

    '''

    nc = model.nc

    if nc != 2:
        raise Exception('3 phase equilibra for binary mixtures')

    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    global vx, vw, vy

    sol1 = root(haz_objb, np.hstack([X0, W0, Y0, P_T]),
                args=(T_P, spec, model, v0))
    error = np.linalg.norm(sol1.fun)
    nfev = sol1.nfev
    sol = sol1.x
    if np.any(sol < 0):
        raise Exception('negative Composition or T/P  founded')
    X, W, Y, var = np.array_split(sol, 4)

    if full_output:
        if spec == 'T':
            P = var
            T = T_P
        elif spec == 'P':
            T = var
            P = T_P
        inc = {'T': T, 'P': P, 'error': error, 'nfev': nfev,
               'X': X, 'vx': vx, 'statex': 'Liquid',
               'W': W, 'vw': vw, 'statew': 'Liquid',
               'Y': Y, 'vy': vy, 'statey': 'Vapor'}
        out = EquilibriumResult(inc)
        return out

    return X, W, Y, var



import numpy as np
from scipy.optimize import root
from ..constants import kb, Na


R = Na * kb


def mu_obj(rho, temp_aux, saft):
    rhol, rhov = Na * rho

    global Xassl, Xassv
    dal, Xassl = saft.d2afcn_aux(rhol, temp_aux, Xassl)
    afcnl, dafcnl, d2afcnl = dal
    Pl = rhol**2 * dafcnl / Na
    dPl = (2 * rhol * dafcnl + rhol**2 * d2afcnl)

    mul = afcnl + rhol*dafcnl
    dmul = Na * (rhol*d2afcnl + 2*dafcnl)

    dav, Xaasv = saft.d2afcn_aux(rhov, temp_aux)
    afcnv, dafcnv, d2afcnv = dav
    Pv = rhov**2 * dafcnv / Na
    dPv = (2 * rhov * dafcnv + rhov**2 * d2afcnv)

    muv = afcnv + rhov * dafcnv
    dmuv = Na * (rhol*d2afcnv + 2*dafcnv)

    FO = np.array([mul-muv, Pl - Pv])
    dFO = np.array([[dmul, -dmuv],
                   [dPl, - dPv]])
    return FO, dFO


def psat(saft, T, P0=None, v0=[None, None], Xass0=[None, None]):

    P0input = P0 is None
    v0input = v0 == [None, None]

    if P0input and v0input:
        print('You need to provide either initial pressure or volumes')
    elif not P0input:
        good_initial = False
        P = P0
    elif not v0input:
        good_initial = True

    temp_aux = saft.temperature_aux(T)
    beta = temp_aux[0]
    RT = Na/beta

    global Xassl, Xassv
    Xassl, Xassv = Xass0

    vl, vv = v0
    if not good_initial:
        lnphiv, vv, Xassv = saft.logfug_aux(temp_aux, P, 'V', vv, Xassv)
        lnphil, vl, Xassv = saft.logfug_aux(temp_aux, P, 'L', vl, Xassl)
        FO = lnphiv - lnphil
        dFO = (vv - vl)/RT
        dP = FO/dFO
        if dP > P:
            dP /= 2
        P -= dP
        for i in range(10):
            lnphiv, vv, Xassv = saft.logfug_aux(temp_aux, P, 'V', vv, Xassv)
            lnphil, vl, Xassv = saft.logfug_aux(temp_aux, P, 'L', vl, Xassl)
            FO = lnphiv - lnphil
            dFO = (vv - vl)/RT
            P -= FO/dFO
            sucess = abs(FO) <= 1e-8
            if sucess:
                break
        if not sucess:
            rho0 = 1. / np.array([vl, vv])
            sol = root(mu_obj, rho0, args=(temp_aux, saft), jac=True)
            rhol, rhov = sol.x
            vl, vv = 1./sol.x
            rhomolecular = rhol * Na
            dal, Xassl = saft.dafcn_aux(rhomolecular, temp_aux, Xassl)
            afcn, dafcn = dal
            P = rhomolecular**2 * dafcn/Na
    else:
        rho0 = 1. / np.asarray([v0])
        sol = root(mu_obj, rho0, args=(temp_aux, saft), jac=True)
        if sol.success:
            rhol, rhov = sol.x
            vl, vv = 1./sol.x
            rhomolecular = rhol * Na
            dal, Xassl = saft.dafcn_aux(rhomolecular, temp_aux, Xassl)
            afcn, dafcn = dal
            P = rhomolecular**2 * dafcn/Na
        else:
            P = None
    return P, vl, vv



import numpy as np


# Hard sphere Eq A6
def ahs(xhi):
    xhi0, xhi1, xhi2, xhi3 = xhi
    a = (xhi2**3/xhi3**2 - xhi0) * np.log(1 - xhi3)
    a += 3 * xhi1 * xhi2 / (1 - xhi3)
    a += xhi2**3 / xhi3 / (1 - xhi3)**2
    a /= xhi0
    return a


def dahs_dxhi00(xhi, dxhi_dxhi00):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00

    a = (xhi2**3/xhi3**2 - xhi0) * np.log(1 - xhi3)
    a += 3 * xhi1 * xhi2 / (1 - xhi3)
    a += xhi2**3 / xhi3 / (1 - xhi3)**2
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * (-1 + xhi3)
    da -= dxhi0 * dxhi3 * (-1 + xhi3)**2
    da /= (-1 + xhi3)**3 * dxhi0

    return np.array([a, da])


def d2ahs_dxhi00(xhi, dxhi_dxhi00):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00

    a = (xhi2**3/xhi3**2 - xhi0) * np.log(1 - xhi3)
    a += 3 * xhi1 * xhi2 / (1 - xhi3)
    a += xhi2**3 / xhi3 / (1 - xhi3)**2
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * (-1 + xhi3)
    da -= dxhi0 * dxhi3 * (-1 + xhi3)**2
    da /= (-1 + xhi3)**3 * dxhi0

    d2a = -6 * dxhi1 * dxhi2 * dxhi3 * (-1 + xhi3)
    d2a += dxhi0 * dxhi3**2 * (-1 + xhi3)**2
    d2a += dxhi2**3 * (3 + xhi3 * (4 - xhi3))
    d2a /= (-1 + xhi3)**4 * dxhi0

    return np.array([a, da, d2a])


def d3ahs_dxhi00(xhi, dxhi_dxhi00):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00

    d3a = -9*dxhi1*dxhi2*dxhi3*(-1+xhi3)
    d3a += dxhi0*dxhi3**2*(-1+xhi3)**2
    d3a += dxhi2**3 * (8 + xhi3 * (5 - xhi3))
    d3a *= 2*dxhi3
    d3a /= -(-1 + xhi3)**5 * dxhi0

    return d3a


def dahs_dx(xhi, dxhi_dx):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dx

    log3 = np.log(1 - xhi3)

    a = (xhi2**3/xhi3**2 - xhi0) * log3
    a += 3 * xhi1 * xhi2 / (1 - xhi3)
    a += xhi2**3 / xhi3 / (1 - xhi3)**2
    a /= xhi0

    dax1 = -3*xhi0*xhi2**2 * (log3 + xhi3/(-1 + xhi3)**2)
    dax1 *= dxhi2 / xhi3**2

    dax2 = (xhi2 * dxhi0 - xhi0 * dxhi2) * (-1 + xhi3)
    dax2 += xhi0 * xhi2 * dxhi3
    dax2 *= -3. * xhi1/(-1 + xhi3)

    dax2 += 3 * xhi0 * xhi2 * dxhi1
    dax2 += xhi0**2 * dxhi3
    dax2 /= (-1 + xhi3)

    dax3 = log3 * dxhi0 - xhi0 * (-2 + xhi3)*dxhi3/(-1 + xhi3)**2
    dax3 *= xhi3
    dax3 += 2*log3*xhi0*dxhi3
    dax3 += xhi3**2 * ((-1+xhi3)*dxhi0 + 2*xhi0*dxhi3)/(-1 + xhi3)**3
    dax3 *= (xhi2/xhi3)**3

    dax = dax1 + dax2 + dax3
    dax /= -xhi0**2

    return a, dax


def dahs_dxxhi(xhi, dxhi_dxhi00, dxhi_dx):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx

    log3 = np.log(1 - xhi3)

    a = (xhi2**3/xhi3**2 - xhi0) * log3
    a += 3 * xhi1 * xhi2 / (1 - xhi3)
    a += xhi2**3 / xhi3 / (1 - xhi3)**2
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * (-1 + xhi3)
    da -= dxhi0 * dxhi3 * (-1 + xhi3)**2
    da /= (-1 + xhi3)**3 * dxhi0

    dax1 = -3*xhi0*xhi2**2 * (log3 + xhi3/(-1 + xhi3)**2)
    dax1 *= dxhi2x / xhi3**2

    dax2 = (xhi2 * dxhi0x - xhi0 * dxhi2x) * (-1 + xhi3)
    dax2 += xhi0 * xhi2 * dxhi3x
    dax2 *= -3. * xhi1/(-1 + xhi3)

    dax2 += 3 * xhi0 * xhi2 * dxhi1x
    dax2 += xhi0**2 * dxhi3x
    dax2 /= (-1 + xhi3)

    dax3 = log3 * dxhi0x - xhi0 * (-2 + xhi3)*dxhi3x/(-1 + xhi3)**2
    dax3 *= xhi3
    dax3 += 2*log3*xhi0*dxhi3x
    dax3 += xhi3**2 * ((-1+xhi3)*dxhi0x + 2*xhi0*dxhi3x)/(-1 + xhi3)**3
    dax3 *= (xhi2/xhi3)**3

    dax = dax1 + dax2 + dax3
    dax /= -xhi0**2

    return np.array([a, da]), dax



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import fsolve
from .multiflash import multiflash
from .equilibriumresult import EquilibriumResult


def haz_objb(inc, T_P, tipo, modelo, index, equilibrio, v0):

    X0 = inc[:-1].reshape(3, 2)
    P_T = inc[-1]

    if tipo == 'T':
        P = P_T
        T = T_P
    elif tipo == 'P':
        T = P_T
        P = T_P

    nc = modelo.nc
    X = np.zeros((3, nc))
    X[:, index] = X0
    lnphi = np.zeros_like(X)

    global vg
    vg = v0.copy()
    for i, estado in enumerate(equilibrio):
        lnphi[i], vg[i] = modelo.logfugef(X[i], T, P, estado, v0[i])

    lnK = lnphi[0] - lnphi[1:]
    K = np.exp(lnK)
    return np.hstack([(K[:, index]*X0[0]-X0[1:]).flatten(), X0.sum(axis=1)-1.])


def haz_pb(X0, P_T, T_P, tipo, modelo, index, equilibrio,
           v0=[None, None, None]):
    sol = fsolve(haz_objb, np.hstack([X0.flatten(), P_T]),
                 args=(T_P, tipo, modelo, index, equilibrio, v0))

    var = sol[-1]
    X = sol[:-1].reshape(3, 2)

    return X, var


def haz_objt(inc, T, P, model, v0=[None, None, None]):

    X, W, Y = np.split(inc, 3)

    global vx, vw, vy
    vx, vw, vy = v0

    fugX, vx = model.logfugef(X, T, P, 'L', vx)
    fugW, vw = model.logfugef(W, T, P, 'L', vw)
    fugY, vy = model.logfugef(Y, T, P, 'V', vy)

    K1 = np.exp(fugX-fugY)
    K2 = np.exp(fugX-fugW)

    return np.hstack([K1*X-Y, K2*X-W, X.sum()-1, Y.sum()-1, W.sum()-1])


def haz(X0, W0, Y0, T, P, model, good_initial=False, v0=[None, None, None],
        Xass0=[None, None, None], K_tol=1e-10, full_output=False):
    """
    Liquid liquid vapor (T,P) -> (x, w, y)

    Computes liquid liquid vapor equilibrium of multicomponent mixtures at
    given temperature. This functions uses a simultaneous method to check
    stability and equilibrium, when slow convergence is noted, minimization of
    Gibbs free energy is performed with BFGS.

    Parameters
    ----------

    X0 : array_like
         guess composition of liquid 1
    W0 : array_like
         guess composition of liquid 2
    Y0 : array_like
         guess composition of vapour 1
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip Gupta's method and solves full system of equations.
    v0 : list, optional
         if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
            Desired accuracy of K (= X/Xr) vector
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector

    """

    nc = model.nc
    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    Z0 = (X0+Y0+W0)/3
    nonzero = np.count_nonzero(Z0)
    x0 = np.vstack([X0, W0, Y0])
    b0 = np.array([0.33, 0.33, 0.33, 0., 0.])

    # check for binary mixture
    if nonzero == 2:
        index = np.nonzero(Z0)[0]
        sol = np.zeros_like(x0)
        global vg
        sol[:, index], T = haz_pb(x0[:, index], T, P, 'P', model,
                                  index, 'LLV', v0)
        X, W, Y = sol
        if full_output:
            info = {'T': T, 'P': P, 'X': sol, 'v': vg,
                    'states': ['L', 'L', 'V']}
            out = EquilibriumResult(info)
            return out
        return X, W, Y, T

    if not good_initial:
        out = multiflash(x0, b0, ['L', 'L', 'V'], Z0, T, P, model, v0, Xass0,
                         K_tol, True)
    else:
        global vx, vw, vy
        sol = fsolve(haz_objt, x0.flatten(), args=(T, P, model, v0))
        x0 = sol.reshape([model.nc, 3])
        Z0 = x0.mean(axis=0)
        out = multiflash(x0, b0, ['L', 'L', 'V'], Z0, T, P, model,
                         [vx, vw, vy], Xass0, K_tol, True)

    Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
    error_inner = out.error_inner
    v = out.v

    if error_inner > 1e-6:
        order = [2, 0, 1]  # Y, X, W
        Xm = Xm[order]
        betatetha = np.hstack([beta[order], tetha])
        equilibrio = np.asarray(equilibrio)[order]
        v0 = np.asarray(v)[order]
        out = multiflash(Xm, betatetha, equilibrio, Z0, T, P, model, v0, Xass0,
                         K_tol, full_output=True)
        order = [1, 2, 0]
        Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
        error_inner = out.error_inner
        if error_inner > 1e-6:
            order = [2, 1, 0]  # W, X, Y
            Xm = Xm[order]
            betatetha = np.hstack([beta[order], tetha])
            equilibrio = np.asarray(equilibrio)[order]
            v0 = np.asarray(out.v)[order]
            out = multiflash(Xm, betatetha, equilibrio, Z0, T, P, model, v0,
                             Xass0, K_tol, full_output=True)
            order = [1, 0, 2]
            Xm, beta, tetha = out.X, out.beta, out.tetha
            equilibrio = out.states
            error_inner = out.error_inner
        Xm = Xm[order]
        beta = beta[order]
        tetha = np.hstack([0., tetha])
        tetha = tetha[order]
        v = (out.v)[order]
    else:
        tetha = np.hstack([0., tetha])

    if full_output:
        info = {'T': T, 'P': P, 'error_outer': out.error_outer,
                'error_inner': error_inner, 'iter': out.iter,
                'beta': beta, 'tetha': tetha, 'X': Xm, 'v': v,
                'states': ['L', 'L', 'V']}
        out = EquilibriumResult(info)
        return out

    tethainestable = tetha > 0.
    Xm[tethainestable] = None
    X, W, Y = Xm

    return X, W, Y


def vlle(X0, W0, Y0, Z, T, P, model, v0=[None, None, None],
         Xass0=[None, None, None], K_tol=1e-10, full_output=False):
    """
    Liquid liquid vapor Multiflash (Z, T,P) -> (x, w, y)

    Computes liquid liquid vapor equilibrium of multicomponent mixtures at
    given temperature. This functions uses a simultaneous method to check
    stability and equilibrium, when slow convergence is noted, minimization
    of Gibbs free energy is performed with BFGS.

    Parameters
    ----------

    X0 : array_like
         guess composition of liquid 1
    W0 : array_like
         guess composition of liquid 2
    Y0 : array_like
         guess composition of vapour 1
    T : float
        absolute temperature in K.
    P : float
        pressure in bar
    model : object
        Created from mixture, eos and mixrule
    good_initial: bool, optional
        if True skip Gupta's method and solves full system of equations.
    v0 : list, optional
         if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= X/Xr) vector
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector

    """

    nc = model.nc
    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    nonzero = np.count_nonzero(Z)
    x0 = np.vstack([X0, W0, Y0])
    b0 = np.array([0.33, 0.33, 0.33, 0., 0.])

    # check for binary mixture
    if nonzero == 2:
        index = np.nonzero(Z)[0]
        sol = np.zeros_like(x0)
        global vg
        sol[:, index], T = haz_pb(x0[:, index], T, P, 'P', model, index,
                                  'LLV', v0)
        X, W, Y = sol
        if full_output:
            info = {'T': T, 'P': P, 'X': sol, 'v': vg,
                    'states': ['L', 'L', 'V']}
            out = EquilibriumResult(info)
            return out
        return X, W, Y, T

    out = multiflash(x0, b0, ['L', 'L', 'V'], Z, T, P, model, v0, Xass0,
                     K_tol, True)

    Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
    error_inner = out.error_inner
    v = out.v
    Xass = out.Xass

    if error_inner > 1e-6:
        order = [2, 0, 1]  # Y, X, W
        Xm = Xm[order]
        betatetha = np.hstack([beta[order], tetha])
        equilibrio = np.asarray(equilibrio)[order]
        v0 = np.asarray(v)[order]
        np.asarray(out.Xass)[order]
        out = multiflash(Xm, betatetha, equilibrio, Z, T, P, model, v0,
                         Xass, K_tol, full_output=True)
        order = [1, 2, 0]
        Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
        error_inner = out.error_inner
        if error_inner > 1e-6:
            order = [2, 1, 0]  # W, X, Y
            Xm = Xm[order]
            betatetha = np.hstack([beta[order], tetha])
            equilibrio = np.asarray(equilibrio)[order]
            v0 = np.asarray(out.v)[order]
            Xass0 = np.asarray(out.Xass)[order]
            out = multiflash(Xm, betatetha, equilibrio, Z, T, P, model, v0,
                             Xass0, K_tol, full_output=True)
            order = [1, 0, 2]
            Xm, beta, tetha = out.X, out.beta, out.tetha
            equilibrio = out.states
            error_inner = out.error_inner
        Xm = Xm[order]
        beta = beta[order]
        tetha = np.hstack([0., tetha])
        tetha = tetha[order]
        v = (out.v)[order]
        Xass = np.asarray(out.Xass)[order]
    else:
        tetha = np.hstack([0., tetha])

    if full_output:
        info = {'T': T, 'P': P, 'error_outer': out.error_outer,
                'error_inner': error_inner, 'iter': out.iter, 'beta': beta,
                'tetha': tetha, 'X': Xm, 'v': v, 'Xass': Xass,
                'states': ['L', 'L', 'V']}
        out = EquilibriumResult(info)
        return out

    tethainestable = tetha > 0.
    Xm[tethainestable] = None
    X, W, Y = Xm

    return X, W, Y


__all__ = ['haz', 'vlle']



import numpy as np


def g2mca(xhi00, khs, xm, da2new, suma_g2, eps, dii):

    cte = 1 / (12 * eps**2 * dii**3)

    g2 = 3 * da2new
    g2 -= eps * khs * suma_g2 / xhi00
    g2 /= xm
    g2 *= cte
    return g2


def dg2mca_dxhi00(xhi00, khs, dkhs, xm, da2new, d2a2new, suma_g2, eps, dii):

    cte = 1 / (12 * eps**2 * dii**3)
    cte2 = eps

    sum1, dsum1 = suma_g2
    g2 = 3*np.array([da2new, d2a2new])

    g2[0] -= cte2 * khs * sum1 / xhi00
    g2[1] += cte2 * khs * sum1 / xhi00**2
    g2[1] -= cte2 * (sum1 * dkhs + khs * dsum1)/xhi00
    g2 /= xm
    g2 *= cte
    return g2


def d2g2mca_dxhi00(xhi00, khs, dkhs, d2khs, xm, da2new, d2a2new, d3a2new,
                   suma_g2, eps, dii):

    cte = 1 / (12 * eps**2 * dii**3)
    cte2 = eps

    sum1, dsum1, d2sum1 = suma_g2

    g2 = 3.*np.array([da2new, d2a2new, d3a2new])

    g2[0] -= cte2 * khs * sum1 / xhi00
    g2[1] += cte2 * khs * sum1 / xhi00**2
    g2[1] -= cte2 * (sum1 * dkhs + khs * dsum1)/xhi00
    g2[2] += 2*cte2*(sum1*dkhs + khs * dsum1)/xhi00**2
    g2[2] -= 2*cte2*khs*sum1/xhi00**3
    g2[2] -= (2*dkhs*dsum1+sum1*d2khs + khs*d2sum1)*cte2/xhi00
    g2 /= xm
    g2 *= cte
    return g2


def dg2mca_dx(xhi00, khs, dkhs, xm, ms, da2new, da2newx, suma_g2, suma_g2x,
              eps, dii):

    cte = 1 / (12 * eps**2 * dii**3)

    g2 = 3 * da2new
    g2 -= eps * khs * suma_g2 / xhi00
    g2 /= xm
    g2 *= cte

    dg2x = -np.multiply.outer(dkhs,  suma_g2)
    dg2x -= khs * suma_g2x
    dg2x += khs * np.multiply.outer(ms, suma_g2)/xm
    dg2x /= xhi00
    dg2x *= eps
    dg2x -= 3 * np.multiply.outer(ms, da2new) / xm
    dg2x += 3. * da2newx
    dg2x /= xm
    dg2x *= cte

    return g2, dg2x


def dg2mca_dxxhi(xhi00, khs, dkhs, dkhsx, xm, ms, da2new, d2a2new, da2newx,
                 suma_g2, suma_g2x, eps, dii):

    cte = 1 / (12 * eps**2 * dii**3)
    cte2 = eps

    sum1, dsum1 = suma_g2
    g2 = 3*np.array([da2new, d2a2new])
    g2[0] -= cte2 * khs * sum1 / xhi00
    g2[1] += cte2 * khs * sum1 / xhi00**2
    g2[1] -= cte2 * (sum1 * dkhs + khs * dsum1)/xhi00
    g2 /= xm
    g2 *= cte

    dg2x = -np.multiply.outer(dkhsx,  suma_g2[0])
    dg2x -= khs * suma_g2x
    dg2x += khs * np.multiply.outer(ms, suma_g2[0])/xm
    dg2x /= xhi00
    dg2x *= eps
    dg2x -= 3 * np.multiply.outer(ms, da2new) / xm
    dg2x += 3. * da2newx
    dg2x /= xm
    dg2x *= cte

    return g2, dg2x



import numpy as np


# Hard sphere Eq 11
def ahs(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    return a


def dahs_deta(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    da = 2*(-2+eta)/(-1+eta)**3
    return np.array([a, da])


def d2ahs_deta(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    da = 2*(-2+eta)/(-1+eta)**3
    d2a = (10-4*eta)/(-1+eta)**4
    return np.array([a, da, d2a])


# Fisr pertubation Eq 34
def a1m(x0lambda, a1sb, c):
    a1 = np.matmul(a1sb, x0lambda)
    a1 *= c
    return a1


# Second pertubation Eq 36
def a2m(x0lambda, a1sb, khs, xi, c, eps):

    sum1 = np.dot(a1sb, x0lambda)
    a2 = khs*(1+xi)*eps*c**2*sum1/2.
    return a2


def da2m_deta(x0lambda, da1sb,  dKhs, dXi, c, eps):

    khs, dkhs = dKhs
    xi, dx1 = dXi
    x1 = 1. + xi

    sum1, dsum1 = np.matmul(da1sb, x0lambda)

    a2 = khs*x1*eps*c**2*sum1/2.

    da2 = sum1*x1*dkhs + khs * x1 * dsum1 + khs * sum1 * dx1
    da2 *= (0.5*eps*c**2)
    return np.hstack([a2, da2])


def d2a2m_deta(x0lambda, d2a1sb,  d2Khs, d2Xi, c, eps):

    khs, dkhs, d2khs = d2Khs
    xi, dx1, d2x1 = d2Xi
    x1 = 1. + xi

    sum1, dsum1, d2sum1 = np.matmul(d2a1sb, x0lambda)

    a2 = khs*x1*eps*c**2*sum1/2.

    da2 = sum1*x1*dkhs + khs * x1 * dsum1 + khs * sum1 * dx1
    da2 *= (0.5*eps*c**2)

    d2a2 = d2khs * sum1 * x1 + d2x1 * sum1 * khs + d2sum1 * khs * x1
    d2a2 += 2 * dkhs * dsum1 * x1
    d2a2 += 2 * sum1 * dkhs * dx1
    d2a2 += 2 * khs * dsum1 * dx1
    d2a2 *= (0.5*eps*c**2)
    return np.hstack([a2, da2, d2a2])


def da2m_new_deta(x0lambda, da1sb,  dKhs, c, eps):

    khs, dkhs = dKhs
    sum1, dsum1 = np.matmul(da1sb, x0lambda)
    da2 = 0.5*eps*c**2*(dkhs*sum1 + khs * dsum1)

    return da2


def d2a2m_new_deta(x0lambda, d2a1sb,  d2Khs, c, eps):

    khs, dkhs, d2khs = d2Khs
    sum1, dsum1, d2sum1 = np.matmul(d2a1sb, x0lambda)

    aux = 0.5*eps*c**2
    da2 = aux*(dkhs*sum1 + khs * dsum1)

    d2a2 = 2 * dkhs * dsum1 + sum1 * d2khs + d2sum1 * khs
    d2a2 *= aux

    return np.hstack([da2, d2a2])


def d3a2m_new_deta(x0lambda, d3a1sb,  d3Khs, c, eps):

    khs, dkhs, d2khs, d3khs = d3Khs
    sum1, dsum1, d2sum1, d3sum1 = np.matmul(d3a1sb, x0lambda)
    aux = 0.5*eps*c**2
    da2 = aux*(dkhs*sum1 + khs * dsum1)

    d2a2 = 2 * dkhs * dsum1 + sum1 * d2khs + d2sum1 * khs
    d2a2 *= aux

    d3a2 = 3 * dsum1 * d2khs + 3 * dkhs * d2sum1
    d3a2 += khs * d3sum1 + d3khs * sum1
    d3a2 *= aux
    return np.hstack([da2, d2a2, d3a2])


# Third pertubation Eq 19
def a3m(x03, nsigma, eps, f4, f5, f6):
    ter1 = -eps**3*f4*nsigma
    ter2 = np.exp(f5*nsigma + f6*nsigma**2)
    return ter1*ter2


def da3m_deta(x03, nsigma, eps, f4, f5, f6):

    aux = np.exp(f5*nsigma + f6*nsigma**2)
    a = -eps**3*f4*nsigma
    a *= aux

    da = -eps**3*f4*aux
    da *= (1. + nsigma * (f5 + 2*f6*nsigma))
    da *= x03
    return np.hstack([a, da])


def d2a3m_deta(x03, nsigma, eps, f4, f5, f6):

    aux = np.exp(f5*nsigma + f6*nsigma**2)
    a = -eps**3*f4*nsigma
    a *= aux

    da = -eps**3*f4*aux
    da *= (1 + nsigma * (f5 + 2*f6*nsigma))
    da *= x03

    d2a = -eps**3*f4*aux
    d2a *= (f5**2*nsigma+2*f6*nsigma*(3+2*f6*nsigma**2)+f5*(2+4*f6*nsigma**2))
    d2a *= x03**2
    return np.hstack([a, da, d2a])




def amono(ahs, a1m, a2m, a3m, beta, ms):
    am = ms * (ahs + beta * a1m + beta**2 * a2m + beta**3 * a3m)
    return am


def damono_drho(ahs, a1m, a2m, a3m, beta, drho, ms):
    am = ms * drho * (ahs + beta * a1m + beta**2 * a2m + beta**3 * a3m)
    return am


def d2amono_drho(ahs, a1m, a2m, a3m, beta, drho, ms):

    am = ms * drho * (ahs + beta * a1m + beta**2 * a2m + beta**3 * a3m)
    return am



import numpy as np


def g1sigma(xhi00, xm, da1, suma_g1, eps, dii):
    cte = 1 / (12 * eps * dii**3)
    g1 = 3 * da1 - suma_g1/xhi00
    g1 /= xm
    g1 *= cte
    return g1


def dg1sigma_dxhi00(xhi00, xm, da1, suma_g1, eps, dii):

    cte = 1 / (12 * eps * dii**3)

    sum1, dsum1 = suma_g1

    g1 = 3 * da1
    g1[0] -= sum1/xhi00
    g1[1] += (sum1/xhi00 - dsum1)/xhi00

    g1 /= xm

    g1 *= cte

    return g1


def d2g1sigma_dxhi00(xhi00, xm, da1, suma_g1, eps, dii):

    cte = 1 / (12 * eps * dii**3)

    sum1, dsum1, d2sum1 = suma_g1
    g1 = 3. * da1 - suma_g1/xhi00
    g1[1] += sum1/xhi00**2
    g1[2] += -2*sum1/xhi00**3 + 2*dsum1/xhi00**2

    g1 /= xm
    g1 *= cte
    return g1


def dg1sigma_dx(xhi00, xm, ms, da1, da1x, suma_g1, suma_g1x, eps, dii):

    cte = 1 / (12 * eps * dii**3)

    g1 = 3. * da1 - suma_g1/xhi00
    g1 /= xm
    g1 *= cte

    g1x = np.multiply.outer(ms, (suma_g1/xhi00 - 3. * da1)/xm)
    g1x += -suma_g1x/xhi00
    g1x += 3*da1x
    g1x /= xm
    g1x *= cte
    return g1, g1x


def dg1sigma_dxxhi(xhi00, xm, ms, da1, da1x, suma_g1, suma_g1x, eps, dii):

    cte = 1 / (12 * eps * dii**3)

    sum1, dsum1 = suma_g1

    g1 = 3 * da1
    g1[0] -= sum1/xhi00
    g1[1] += (sum1/xhi00 - dsum1)/xhi00

    g1 /= xm
    g1 *= cte

    g1x = np.multiply.outer(ms, (sum1/xhi00 - 3. * da1[0])/xm)
    g1x += -suma_g1x/xhi00
    g1x += 3*da1x
    g1x /= xm
    g1x *= cte
    return g1, g1x



from .gmie import lngmie, dlngmie_drho, d2lngmie_drho


def achain(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
           x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho, beta, eps,
           c, ms, ring):
    lng = lngmie(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
                 x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho, beta,
                 eps, c, ms)
    ac = - (ms - 1. + ring*eta)*lng
    return ac


def dachain_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha, tetha,
                 x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho, beta,
                 eps, c, ms, ring):
    dlng = dlngmie_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha,
                        tetha, x0_a2, a1sb_a2, da2m_new_deta, khs, d,
                        deta_drho, rho, beta, eps, c, ms)
    dac = - (ms - 1. + ring*eta)*dlng
    return dac


def d2achain_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma, alpha,
                  tetha, x0_a2, a1sb_a2, da2m_new_deta, khs, d, deta_drho, rho,
                  beta, eps, c, ms, ring):
    d2lng = d2lngmie_drho(x0, eta, x0_a1, a1sb_a1, da1m_deta, x03, nsigma,
                          alpha, tetha, x0_a2, a1sb_a2, da2m_new_deta, khs,
                          d, deta_drho, rho, beta, eps, c, ms)
    d2ac = - (ms - 1. + ring*eta)*d2lng
    return d2ac



import numpy as np
from scipy.optimize import fsolve


def association_config(sitesmix, eos):

    compindex = []
    for i, j in enumerate(sitesmix):
        compindex += np.count_nonzero(j) * [i]
    compindex = np.asarray(compindex)

    types = np.array([['B', 'P', 'N']])
    n = len(sitesmix)
    n = eos.nc
    types = np.repeat(types, n, axis=0)
    nozero = np.nonzero(sitesmix)
    types = types[nozero]
    ntypes = np.asarray(sitesmix)[nozero]
    nsites = len(types)

    S = np.array(sitesmix)
    S = S[S != 0]

    DIJ = np.zeros([nsites, nsites])
    int_i = []
    int_j = []
    for i in range(nsites):
        for j in range(nsites):
            bool1 = types[i] == 'B'
            bool2 = types[i] == 'P' and (types[j] == 'N' or types[j] == 'B')
            bool3 = types[i] == 'N' and (types[j] == 'P' or types[j] == 'B')
            if bool1 or bool2 or bool3:
                DIJ[i, j] = ntypes[j]
                int_i.append(i)
                int_j.append(j)

    indexabij = (int_i, int_j)
    indexab = (compindex[int_i], compindex[int_j])
    diagasso = np.diag_indices(nsites)

    dxjdx = np.zeros([eos.nc, nsites])
    for i in range(eos.nc):
        dxjdx[i] = compindex == i

    return S, DIJ, compindex, indexabij, indexab, nsites, dxjdx, diagasso


def Iab(xhi, dii, dij, rcij, rdij, sigmaij3):

    xhi0, xhi1, xhi2, xhi3 = xhi
    aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)

    xhi13 = 1 - xhi3

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13**2
    gdhs += 2 * aux**2 * xhi2**2 / xhi13**3

    Kab = np.log((rcij + 2*rdij)/dij)
    Kab *= 6*rcij**3 + 18 * rcij**2*rdij - 24 * rdij**3
    aux1 = (rcij + 2 * rdij - dij)
    aux2 = (22*rdij**2 - 5*rcij*rdij - 7*rdij*dij - 8*rcij**2+rcij*dij+dij**2)
    Kab += aux1 * aux2
    Kab /= (72*rdij**2 * sigmaij3)
    Kab *= 4 * np.pi * dij**2

    iab = gdhs * Kab
    return iab


def dIab_drho(xhi, dxhi_dxhi00, dxhi00_drho, dii, dij, rcij, rdij, sigmaij3):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)

    xhi13 = 1 - xhi3

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13**2
    gdhs += 2 * aux**2 * xhi2**2 / xhi13**3

    dgdhs = 4. * aux**2 * xhi2 * dxhi2 / xhi13**3
    dgdhs += 3. * aux * dxhi2 / xhi13**2
    dgdhs += 6. * aux**2 * xhi2**2 * dxhi3 / xhi13**4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13**3
    dgdhs += dxhi3 / xhi13**2
    dgdhs *= dxhi00_drho

    Kab = np.log((rcij + 2*rdij)/dij)
    Kab *= 6*rcij**3 + 18 * rcij**2*rdij - 24 * rdij**3
    aux1 = (rcij + 2 * rdij - dij)
    aux2 = (22*rdij**2 - 5*rcij*rdij - 7*rdij*dij - 8*rcij**2+rcij*dij+dij**2)
    Kab += aux1 * aux2
    Kab /= (72*rdij**2 * sigmaij3)
    Kab *= 4 * np.pi * dij**2
    iab = gdhs * Kab
    diab = dgdhs * Kab
    return iab, diab


def d2Iab_drho(xhi, dxhi_dxhi00, dxhi00_drho, dii, dij, rcij, rdij, sigmaij3):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)

    xhi13 = 1 - xhi3

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13**2
    gdhs += 2 * aux**2 * xhi2**2 / xhi13**3

    dgdhs = 4. * aux**2 * xhi2 * dxhi2 / xhi13**3
    dgdhs += 3. * aux * dxhi2 / xhi13**2
    dgdhs += 6. * aux**2 * xhi2**2 * dxhi3 / xhi13**4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13**3
    dgdhs += dxhi3 / xhi13**2
    dgdhs *= dxhi00_drho

    d2gdhs = 2 * dxhi3**2 / xhi13**3
    d2gdhs += 2 * aux**2 * 2 * dxhi2**2 / xhi13**3
    d2gdhs += 2 * aux**2 * 12 * xhi2 * dxhi2 * dxhi3 / xhi13**4
    d2gdhs += 2 * aux**2 * 12 * xhi2**2 * dxhi3**2 / xhi13**5
    d2gdhs += 3 * aux * 4 * dxhi2 * dxhi3 / xhi13**3
    d2gdhs += 3 * aux * 6 * xhi2 * dxhi3**2 / xhi13**4
    d2gdhs *= dxhi00_drho**2

    Kab = np.log((rcij + 2*rdij)/dij)
    Kab *= 6*rcij**3 + 18 * rcij**2*rdij - 24 * rdij**3
    aux1 = (rcij + 2 * rdij - dij)
    aux2 = (22*rdij**2 - 5*rcij*rdij - 7*rdij*dij-8*rcij**2+rcij*dij+dij**2)
    Kab += aux1 * aux2
    Kab /= (72*rdij**2 * sigmaij3)
    Kab *= 4 * np.pi * dij**2

    iab = gdhs * Kab
    diab = dgdhs * Kab
    d2iab = d2gdhs * Kab

    return iab, diab, d2iab


def dIab_dx(xhi, dxhi_dx, dii, dij, rcij, rdij, sigmaij3):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx

    aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)

    xhi13 = 1 - xhi3

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13**2
    gdhs += 2 * aux**2 * xhi2**2 / xhi13**3

    dgdhsx = 4 * np.multiply.outer(aux**2, dxhi2x) * xhi2 / (1-xhi3)**3
    dgdhsx += 3 * np.multiply.outer(aux, dxhi2x) / (1-xhi3)**2
    dgdhsx += 6 * np.multiply.outer(aux**2, dxhi3x) * xhi2**2 / (1-xhi3)**4
    dgdhsx += 6 * np.multiply.outer(aux, dxhi3x) * xhi2 / (1-xhi3)**3
    dgdhsx += dxhi3x / (1-xhi3)**2
    dgdhsx = dgdhsx.T

    Kab = np.log((rcij + 2*rdij)/dij)
    Kab *= 6*rcij**3 + 18 * rcij**2*rdij - 24 * rdij**3
    aux1 = (rcij + 2 * rdij - dij)
    aux2 = (22*rdij**2 - 5*rcij*rdij - 7*rdij*dij - 8*rcij**2+rcij*dij+dij**2)
    Kab += aux1 * aux2
    Kab /= (72*rdij**2 * sigmaij3)
    Kab *= 4 * np.pi * dij**2

    iab = gdhs * Kab
    diabx = dgdhsx * Kab
    return iab, diabx


def dIab_dxrho(xhi, dxhi_dxhi00, dxhi00_drho, dxhi_dx, dii, dij, rcij, rdij,
               sigmaij3):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx
    aux = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)

    xhi13 = 1 - xhi3

    gdhs = 1 / xhi13
    gdhs += 3 * aux * xhi2 / xhi13**2
    gdhs += 2 * aux**2 * xhi2**2 / xhi13**3

    dgdhs = 4. * aux**2 * xhi2 * dxhi2 / xhi13**3
    dgdhs += 3. * aux * dxhi2 / xhi13**2
    dgdhs += 6. * aux**2 * xhi2**2 * dxhi3 / xhi13**4
    dgdhs += 6. * aux * xhi2 * dxhi3 / xhi13**3
    dgdhs += dxhi3 / xhi13**2
    dgdhs *= dxhi00_drho

    dgdhsx = 4 * np.multiply.outer(aux**2, dxhi2x) * xhi2 / (1-xhi3)**3
    dgdhsx += 3 * np.multiply.outer(aux, dxhi2x) / (1-xhi3)**2
    dgdhsx += 6 * np.multiply.outer(aux**2, dxhi3x) * xhi2**2 / (1-xhi3)**4
    dgdhsx += 6 * np.multiply.outer(aux, dxhi3x) * xhi2 / (1-xhi3)**3
    dgdhsx += dxhi3x / (1-xhi3)**2
    dgdhsx = dgdhsx.T

    Kab = np.log((rcij + 2*rdij)/dij)
    Kab *= 6*rcij**3 + 18 * rcij**2*rdij - 24 * rdij**3
    aux1 = (rcij + 2 * rdij - dij)
    aux2 = (22*rdij**2 - 5*rcij*rdij - 7*rdij*dij - 8*rcij**2+rcij*dij+dij**2)
    Kab += aux1 * aux2
    Kab /= (72*rdij**2 * sigmaij3)
    Kab *= 4 * np.pi * dij**2

    iab = gdhs * Kab
    diab = dgdhs * Kab
    diabx = dgdhsx * Kab
    return iab, diab, diabx


def fobj_xass(Xass, xj, aux_asso, diagasso):
    fo = Xass - 1 / (1 + aux_asso@(xj*Xass))
    return fo


def fobj_xass_jac(Xass, xj, aux_asso, diagasso):
    den = 1 + aux_asso@(xj*Xass)
    dfo = ((aux_asso*xj).T/den**2).T
    dfo[diagasso] += 1.
    return dfo


def Xass_solver(nsites, xj, rho, DIJ, Dabij, diagasso, Xass0):

    Xass = 1. * Xass0

    aux_asso = rho * DIJ * Dabij

    omega = 0.2
    for i in range(5):
        den = 1. + aux_asso@(xj*Xass)
        fo = 1. / den
        dXass = (1 - omega) * (fo - Xass)
        Xass += dXass

    bool_method = not np.any(xj == 0.)

    if bool_method:
        KIJ = np.outer(xj, xj) * aux_asso

        KIJXass = KIJ@Xass
        dQ = xj * (1/Xass - 1) - KIJXass
        HIJ = -1 * KIJ
        HIJ[diagasso] -= (xj + KIJXass)/Xass
        for i in range(15):
            dXass = np.linalg.solve(HIJ, -dQ)
            Xnew = Xass + dXass
            Xnew_neg = Xnew < 0
            Xnew[Xnew_neg] = 0.2*Xass[Xnew_neg]
            Xass = Xnew
            KIJXass = KIJ@Xass
            dQ = xj * (1/Xass - 1) - KIJXass
            sucess = np.linalg.norm(dQ) < 1e-8
            if sucess:
                break
            HIJ = -1 * KIJ
            HIJ[diagasso] -= (xj + KIJXass)/Xass
        '''
        for i in range(15):
            KIJXass = KIJ@Xass
            dQ = xj * (1/Xass - 1) - KIJXass
            HIJ = -1 * KIJ
            HIJ[diagasso] -= (xj + KIJXass)/Xass
            dXass = np.linalg.solve(HIJ, -dQ)
            Xass += dXass
            sucess = np.linalg.norm(dXass) < 1e-8
            if sucess:
                break
         '''
    else:
        Xass = fsolve(fobj_xass, x0=Xass, args=(xj, aux_asso, diagasso),
                      fprime=fobj_xass_jac)

    return Xass


def CIJ_matrix(rhom, xj, Xass, DIJ, Dabij, diagasso):
    CIJ = rhom * np.outer(Xass**2, xj) * Dabij * DIJ
    CIJ[diagasso] += 1.
    return CIJ


def dXass_drho(rhom, xj, Xass, DIJ, Dabij, dDabij_drho, CIJ):
    brho = -(DIJ*(Dabij + rhom * dDabij_drho))@(xj * Xass)
    brho *= Xass**2
    dXass = np.linalg.solve(CIJ, brho)
    return dXass


def d2Xass_drho(rhom, xj, Xass, dXass_drho, DIJ, Dabij, dDabij_drho,
                d2Dabij_drho, CIJ):

    b2rho = -rhom * (DIJ * (Xass * d2Dabij_drho+2*dXass_drho*dDabij_drho))@xj
    b2rho += 2 * (1/Xass - 1) / (rhom**2)
    b2rho *= Xass**2
    b2rho += 2 * dXass_drho / (rhom)
    b2rho += 2 * dXass_drho**2 / (Xass)

    d2Xass = np.linalg.solve(CIJ, b2rho)
    return d2Xass


def dXass_dx(rhom, xj, Xass, DIJ, Dabij, dDabij_dx, dxjdx, CIJ):

    bx = (dxjdx * Xass)@(DIJ*Dabij).T + (DIJ * dDabij_dx)@(xj * Xass)
    bx *= Xass**2
    bx *= -rhom
    dXass = np.linalg.solve(CIJ, bx.T)

    return dXass.T



import numpy as np
from .monomer_aux import I_lam, J_lam


# Equation A12
def B(xhi00, xhix, xm, x0, lam, a1vdw_cte):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*I/xhix13 - 9.*xhix*(1+xhix)*J/(2.*xhix13)
    b *= -a1vdw_cte * xm * xhi00
    return b


def dB_dxhi00(xhi00, xhix, xm, x0,  lam, a1vdw_cte, dxhix_dxhi00):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*I/xhix13 - 9.*xhix*(1+xhix)*J/(2.*xhix13)
    b *= -a1vdw_cte * xm * xhi00

    db = (-1 + xhix) * (-2*I + xhix * (I + 9*J + 9*J*xhix))
    db -= xhi00 * dxhix_dxhi00 * (-5*I+9*J+xhix*(2*I+36*J+9*J*xhix))
    db *= -a1vdw_cte * xm
    db /= 2 * (-1 + xhix)**4

    return b, db


def d2B_dxhi00(xhi00, xhix, xm, x0, lam, a1vdw_cte, dxhix_dxhi00):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*I/xhix13 - 9.*xhix*(1+xhix)*J/(2.*xhix13)
    b *= -a1vdw_cte * xm * xhi00

    db = (-1 + xhix) * (-2*I+xhix*(I+9*J+9*J*xhix))
    db -= xhi00 * dxhix_dxhi00 * (-5*I+9*J+xhix*(2*I+36*J+9*J*xhix))
    db *= -a1vdw_cte * xm

    db /= 2 * (-1 + xhix)**4

    d2b = -2*(-1+xhix) * (-5*I+9*J+xhix*(2*I+36*J+9*J*xhix)) * dxhix_dxhi00
    d2b += 6 * xhi00 * (-3*(I-4*J) + xhix * (I+21*J+3*J*xhix))*dxhix_dxhi00**2
    d2b *= -a1vdw_cte * xm
    d2b /= 2 * (-1 + xhix)**5
    return b, db, d2b


def d3B_dxhi00(xhi00, xhix, xm, x0, lam, a1vdw_cte, dxhix_dxhi00):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*I/xhix13 - 9.*xhix*(1+xhix)*J/(2.*xhix13)

    b *= -a1vdw_cte * xm * xhi00

    db = (-1 + xhix) * (-2*I+xhix*(I+9*J+9*J*xhix))
    db -= xhi00 * dxhix_dxhi00 * (-5*I+9*J+xhix*(2*I+36*J+9*J*xhix))
    db *= -a1vdw_cte * xm
    db /= 2 * (-1 + xhix)**4

    d2b = -2*(-1 + xhix) * (-5*I+9*J+xhix*(2*I+36*J+9*J*xhix)) * dxhix_dxhi00
    d2b += 6 * xhi00 * (-3*(I-4*J) + xhix*(I+21*J+3*J*xhix)) * dxhix_dxhi00**2
    d2b *= -a1vdw_cte * xm

    d2b /= 2 * (-1 + xhix)**5

    d3b = -18*(-1 + xhix)*(-3*(I-4*J)+xhix*(I+21*J+3*J*xhix))*dxhix_dxhi00**2
    d3b += 6*xhi00*(-14*I+81*J+xhix*(4*I+90*J+9*J*xhix)) * dxhix_dxhi00**3
    d3b *= a1vdw_cte * xm

    d3b /= 2 * (-1 + xhix)**6
    return b, db, d2b, d3b


def dB_dx(xhi00, xhix, xm, ms, x0, lam, a1vdw_cte, dxhix_dx):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    b = np.multiply.outer((-1+xhix)*(-2*I+xhix*(I+9*J+9*J*xhix)), ms).T
    b += np.multiply.outer(-(-5*I+9*J+xhix*(2*I+36*J+9*J*xhix))*xm, dxhix_dx).T
    b *= - a1vdw_cte * xhi00

    b /= 2 * (-1 + xhix)**4
    return b


def dB_dx_dxhi00(xhi00, xhix, xm, ms, x0, lam, a1vdw_cte, dxhix_dxhi00,
                 dxhix_dx, dxhix_dx_dxhi00):
    I = I_lam(x0, lam)
    J = J_lam(x0, lam)

    aux1 = 2*I + xhix * (-3*(I+3*J) + xhix * (I + 9*J*xhix))
    aux1 += -xhi00 * (-5*I + 9*J + xhix * (2*I+36*J+9*J*xhix)) * dxhix_dxhi00
    aux1 *= (-1 + xhix)

    aux2 = - 5*I + 9*J + xhix * (2*I + 36*J + 9*J*xhix)
    aux2 *= -(-1+xhix)
    aux2 += 6*xhi00 * (-3*(I-4*J)+xhix*(I+21*J+3*J*xhix))*dxhix_dxhi00
    aux2 *= xm

    aux3 = - 5 * I + 9*J + xhix * (2*I + 36*J + 9*J*xhix)
    aux3 *= -xhi00 * (-1 + xhix)
    aux3 *= xm

    b = np.multiply.outer(ms, aux1)
    b += np.multiply.outer(dxhix_dx, aux2)
    b += np.multiply.outer(dxhix_dx_dxhi00, aux3)

    b *= -a1vdw_cte
    b /= (2 * (-1 + xhix)**5)
    return b



import numpy as np
from .a1s_monomer import a1s, da1s_dxhi00, d2a1s_dxhi00, d3a1s_dxhi00
from .a1s_monomer import da1s_dx, da1s_dx_dxhi00
from .B_monomer import B, dB_dxhi00, d2B_dxhi00, d3B_dxhi00, dB_dx
from .B_monomer import dB_dx_dxhi00


def a1sB(xhi00, xhix, xm, x0, lam, cictes, a1vdw, a1vdw_cte):
    a1 = a1s(xhi00, xhix, xm, cictes, a1vdw)
    b = B(xhi00, xhix, xm, x0, lam, a1vdw_cte)
    return a1 + b


def da1sB_dxhi00(xhi00, xhix, xm, x0, lam, cictes, a1vdw, a1vdw_cte,
                 dxhix_dxhi00):

    a1, da1 = da1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw, dxhix_dxhi00)
    b, db = dB_dxhi00(xhi00, xhix, xm, x0,  lam, a1vdw_cte, dxhix_dxhi00)

    return a1 + b, da1 + db


def d2a1sB_dxhi00(xhi00, xhix, xm, x0, lam, cictes,
                  a1vdw, a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1 = d2a1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw, dxhix_dxhi00)
    b, db, d2b = d2B_dxhi00(xhi00, xhix, xm, x0,  lam, a1vdw_cte, dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b


def d3a1sB_dxhi00(xhi00, xhix, xm, x0, lam, cictes, a1vdw, a1vdw_cte,
                  dxhix_dxhi00):
    a1, da1, d2a1, d3a1 = d3a1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw,
                                       dxhix_dxhi00)
    b, db, d2b, d3b = d3B_dxhi00(xhi00, xhix, xm, x0, lam, a1vdw_cte,
                                 dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b, d3a1 + d3b


def da1sB_dx(xhi00, xhix, xm, ms, x0, lam, cictes, a1vdw, a1vdw_cte, dxhix_dx):
    a1 = da1s_dx(xhi00, xhix, xm, ms, cictes, a1vdw, dxhix_dx)
    b = dB_dx(xhi00, xhix, xm, ms, x0, lam, a1vdw_cte,  dxhix_dx)
    return a1 + b


def da1sB_dx_dxhi00(xhi00, xhix, xm, ms, x0, lam, cictes, a1vdw, a1vdw_cte,
                    dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):
    a1 = da1s_dx_dxhi00(xhi00, xhix, xm, ms, cictes, a1vdw, dxhix_dxhi00,
                        dxhix_dx,  dxhix_dx_dxhi00)
    b = dB_dx_dxhi00(xhi00, xhix, xm, ms, x0, lam, a1vdw_cte, dxhix_dxhi00,
                     dxhix_dx, dxhix_dx_dxhi00)
    return a1 + b


def a1sB_eval(xhi00, xhix, x0, xm, lambdas, cctes, a1vdw, a1vdw_cte):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a = a1sB(xhi00, xhix, xm, x0,  laij, cctes_laij, a1vdw_laij,
                  a1vdw_cte)
    a1sb_r = a1sB(xhi00, xhix, xm, x0,  lrij, cctes_lrij, a1vdw_lrij,
                  a1vdw_cte)
    a1sb_2a = a1sB(xhi00, xhix, xm, x0,  2*laij, cctes_2laij, a1vdw_2laij,
                   a1vdw_cte)
    a1sb_2r = a1sB(xhi00, xhix, xm, x0,  2*lrij, cctes_2lrij, a1vdw_2lrij,
                   a1vdw_cte)
    a1sb_ar = a1sB(xhi00, xhix, xm, x0,  larij, cctes_larij, a1vdw_larij,
                   a1vdw_cte)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])
    return a1sb_a1, a1sb_a2


def da1sB_dxhi00_eval(xhi00, xhix, x0, xm, lambdas, cctes, a1vdw, a1vdw_cte,
                      dxhix_dxhi00):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a, da1sb_a = da1sB_dxhi00(xhi00, xhix, xm,  x0, laij, cctes_laij,
                                   a1vdw_laij, a1vdw_cte, dxhix_dxhi00)

    a1sb_r, da1sb_r = da1sB_dxhi00(xhi00, xhix, xm,  x0, lrij, cctes_lrij,
                                   a1vdw_lrij, a1vdw_cte, dxhix_dxhi00)

    a1sb_2a, da1sb_2a = da1sB_dxhi00(xhi00, xhix, xm,  x0, 2*laij, cctes_2laij,
                                     a1vdw_2laij, a1vdw_cte, dxhix_dxhi00)

    a1sb_2r, da1sb_2r = da1sB_dxhi00(xhi00, xhix, xm,  x0, 2*lrij, cctes_2lrij,
                                     a1vdw_2lrij, a1vdw_cte, dxhix_dxhi00)

    a1sb_ar, da1sb_ar = da1sB_dxhi00(xhi00, xhix, xm, x0, larij, cctes_larij,
                                     a1vdw_larij, a1vdw_cte, dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])
    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    return a1sb_a1, a1sb_a2


def d2a1sB_dxhi00_eval(xhi00, xhix, x0, xm, lambdas, cctes, a1vdw, a1vdw_cte,
                       dxhix_dxhi00):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a, da1sb_a, d2a1sb_a = d2a1sB_dxhi00(xhi00, xhix, xm, x0, laij,
                                              cctes_laij, a1vdw_laij,
                                              a1vdw_cte, dxhix_dxhi00)

    a1sb_r, da1sb_r, d2a1sb_r = d2a1sB_dxhi00(xhi00, xhix, xm,  x0, lrij,
                                              cctes_lrij, a1vdw_lrij,
                                              a1vdw_cte, dxhix_dxhi00)

    a1sb_2a, da1sb_2a, d2a1sb_2a = d2a1sB_dxhi00(xhi00, xhix, xm,  x0, 2*laij,
                                                 cctes_2laij, a1vdw_2laij,
                                                 a1vdw_cte, dxhix_dxhi00)

    a1sb_2r, da1sb_2r, d2a1sb_2r = d2a1sB_dxhi00(xhi00, xhix, xm, x0, 2*lrij,
                                                 cctes_2lrij, a1vdw_2lrij,
                                                 a1vdw_cte, dxhix_dxhi00)

    a1sb_ar, da1sb_ar, d2a1sb_ar = d2a1sB_dxhi00(xhi00, xhix, xm, x0, larij,
                                                 cctes_larij, a1vdw_larij,
                                                 a1vdw_cte, dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])
    return a1sb_a1, a1sb_a2


def d3a1sB_dxhi00_eval(xhi00, xhix, x0, xm, lambdas, cctes, a1vdw, a1vdw_cte,
                       dxhix_dxhi00):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a, da1sb_a, d2a1sb_a, d3a1sb_a = d3a1sB_dxhi00(xhi00, xhix, xm,  x0,
                                                        laij, cctes_laij,
                                                        a1vdw_laij, a1vdw_cte,
                                                        dxhix_dxhi00)

    a1sb_r, da1sb_r, d2a1sb_r, d3a1sb_r = d3a1sB_dxhi00(xhi00, xhix, xm, x0,
                                                        lrij, cctes_lrij,
                                                        a1vdw_lrij, a1vdw_cte,
                                                        dxhix_dxhi00)

    a1sb_2a, da1sb_2a, d2a1sb_2a, d3a1sb_2a = d3a1sB_dxhi00(xhi00, xhix, xm,
                                                            x0, 2*laij,
                                                            cctes_2laij,
                                                            a1vdw_2laij,
                                                            a1vdw_cte,
                                                            dxhix_dxhi00)

    a1sb_2r, da1sb_2r, d2a1sb_2r, d3a1sb_2r = d3a1sB_dxhi00(xhi00, xhix, xm,
                                                            x0, 2*lrij,
                                                            cctes_2lrij,
                                                            a1vdw_2lrij,
                                                            a1vdw_cte,
                                                            dxhix_dxhi00)

    a1sb_ar, da1sb_ar, d2a1sb_ar, d3a1sb_ar = d3a1sB_dxhi00(xhi00, xhix, xm,
                                                            x0, larij,
                                                            cctes_larij,
                                                            a1vdw_larij,
                                                            a1vdw_cte,
                                                            dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r],
                       [d3a1sb_a, d3a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r],
                       [d3a1sb_2a, d3a1sb_ar, d3a1sb_2r]])

    return a1sb_a1, a1sb_a2


def da1sB_dx_eval(xhi00, xhix, x0, xm, ms, lambdas, cctes, a1vdw, a1vdw_cte,
                  dxhix_dx):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a = da1sB_dx(xhi00, xhix, xm, ms,  x0, laij, cctes_laij, a1vdw_laij,
                      a1vdw_cte, dxhix_dx)

    a1sb_r = da1sB_dx(xhi00, xhix, xm, ms, x0, lrij, cctes_lrij, a1vdw_lrij,
                      a1vdw_cte, dxhix_dx)

    a1sb_2a = da1sB_dx(xhi00, xhix, xm, ms, x0, 2*laij, cctes_2laij,
                       a1vdw_2laij, a1vdw_cte, dxhix_dx)

    a1sb_2r = da1sB_dx(xhi00, xhix, xm, ms, x0, 2*lrij, cctes_2lrij,
                       a1vdw_2lrij, a1vdw_cte, dxhix_dx)

    a1sb_ar = da1sB_dx(xhi00, xhix, xm, ms, x0, larij, cctes_larij,
                       a1vdw_larij, a1vdw_cte, dxhix_dx)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])

    return a1sb_a1, a1sb_a2


def da1sB_dx_dxhi00_eval(xhi00, xhix, x0, xm, ms, lambdas, cctes, a1vdw,
                         a1vdw_cte, dxhix_dxhi00,  dxhix_dx, dxhix_dx_dxhi00):

    laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctes
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdw

    a1sb_a = da1sB_dx_dxhi00(xhi00, xhix, xm, ms,  x0, laij, cctes_laij,
                             a1vdw_laij, a1vdw_cte, dxhix_dxhi00,
                             dxhix_dx,  dxhix_dx_dxhi00)

    a1sb_r = da1sB_dx_dxhi00(xhi00, xhix, xm, ms, x0, lrij, cctes_lrij,
                             a1vdw_lrij, a1vdw_cte, dxhix_dxhi00,
                             dxhix_dx,  dxhix_dx_dxhi00)

    a1sb_2a = da1sB_dx_dxhi00(xhi00, xhix, xm, ms, x0, 2*laij, cctes_2laij,
                              a1vdw_2laij, a1vdw_cte, dxhix_dxhi00,
                              dxhix_dx,  dxhix_dx_dxhi00)

    a1sb_2r = da1sB_dx_dxhi00(xhi00, xhix, xm, ms, x0, 2*lrij, cctes_2lrij,
                              a1vdw_2lrij, a1vdw_cte, dxhix_dxhi00,
                              dxhix_dx,  dxhix_dx_dxhi00)

    a1sb_ar = da1sB_dx_dxhi00(xhi00, xhix, xm, ms, x0, larij, cctes_larij,
                              a1vdw_larij, a1vdw_cte, dxhix_dxhi00,
                              dxhix_dx, dxhix_dx_dxhi00)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])

    return a1sb_a1, a1sb_a2


def x0lambda_eval(x0, la, lr, lar, laij, lrij, larij, diag_index):
    x0la = x0**laij
    x0lr = x0**lrij
    x02la = x0**(2*laij)
    x02lr = x0**(2*lrij)
    x0lar = x0**larij

    # To be used for a1 and a2 og monomer
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])

    # To be used in g1 and g2 of chain
    x0_g1 = np.array([la * x0la[diag_index], -lr*x0lr[diag_index]])
    x0_g2 = np.array([la * x02la[diag_index], -lar*x0lar[diag_index],
                     lr * x02lr[diag_index]])

    return x0_a1, x0_a2, x0_g1, x0_g2



from __future__ import division, print_function, absolute_import
import numpy as np
from ..math import gdem
from scipy.optimize import minimize
from .equilibriumresult import EquilibriumResult


def rachfordrice(beta, K, Z):
    '''
    Solves Rachford Rice equation by Halley's method
    '''
    K1 = K-1.
    g0 = np.dot(Z, K) - 1.
    g1 = 1. - np.dot(Z, 1/K)
    singlephase = False

    if g0 < 0:
        beta = 0.
        D = np.ones_like(Z)
        singlephase = True
    elif g1 > 0:
        beta = 1.
        D = 1 + K1
        singlephase = True
    it = 0
    e = 1.
    while e > 1e-8 and it < 20 and not singlephase:
        it += 1
        D = 1 + beta*K1
        KD = K1/D
        fo = np.dot(Z, KD)
        dfo = - np.dot(Z, KD**2)
        d2fo = 2*np.dot(Z, KD**3)
        dbeta = - (2*fo*dfo)/(2*dfo**2-fo*d2fo)
        beta += dbeta
        e = np.abs(dbeta)

    return beta, D, singlephase


def Gibbs_obj(v, phases, Z, temp_aux, P, model):
    '''
    Objective function to minimize Gibbs energy in biphasic flash
    '''
    l = Z-v
    v[v < 1e-8] = 1e-8
    l[l < 1e-8] = 1e-8
    X = l/l.sum()
    Y = v/v.sum()
    global v1, v2, Xass1, Xass2
    lnfugl, v1, Xass1 = model.logfugef_aux(X, temp_aux, P, phases[0], v1,
                                           Xass1)
    lnfugv, v2, Xass2 = model.logfugef_aux(Y, temp_aux, P, phases[1], v2,
                                           Xass2)
    fugl = np.log(X) + lnfugl
    fugv = np.log(Y) + lnfugv
    fo = v*fugv + l*fugl
    f = np.sum(fo)
    df = fugv - fugl
    return f, df


def flash(x_guess, y_guess, equilibrium, Z, T, P, model, v0=[None, None],
          Xass0=[None, None], K_tol=1e-8, full_output=False):
    """
    Isothermic isobaric flash (z, T, P) -> (x,y,beta)

    Parameters
    ----------

    x_guess : array_like
        guess composition of phase 1
    y_guess : array_like
        guess composition of phase 2
    equilibrium : string
        'LL' for LLE, 'LV' for VLE
    z : array_like
        overall system composition
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= Y/X) vector
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        phase 1 composition
    Y : array_like
        phase 2 composition
    beta : float
        phase 2 phase fraction
    """
    nc = model.nc
    if len(x_guess) != nc or len(y_guess) != nc or len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    temp_aux = model.temperature_aux(T)
    v10, v20 = v0
    Xass10, Xass20 = Xass0
    e1 = 1
    itacc = 0
    it = 0
    it2 = 0
    n = 4
    nacc = 3
    X = x_guess
    Y = y_guess
    global v1, v2, Xass1, Xass2
    fugl, v1, Xass1 = model.logfugef_aux(X, temp_aux, P, equilibrium[0], v10,
                                         Xass10)
    fugv, v2, Xass2 = model.logfugef_aux(Y, temp_aux, P, equilibrium[1], v20,
                                         Xass20)
    lnK = fugl - fugv
    K = np.exp(lnK)

    bmin = max(np.hstack([((K*Z-1.)/(K-1.))[K > 1], 0.]))
    bmax = min(np.hstack([((1.-Z)/(1.-K))[K < 1], 1.]))
    beta = (bmin + bmax)/2

    while e1 > K_tol and itacc < nacc:
        it += 1
        it2 += 1
        lnK_old = lnK
        beta, D, singlephase = rachfordrice(beta, K, Z)

        X = Z/D
        Y = X*K
        X /= X.sum()
        Y /= Y.sum()
        fugl, v1, Xass1 = model.logfugef_aux(X, temp_aux, P, equilibrium[0],
                                             v1, Xass1)
        fugv, v2, Xass2 = model.logfugef_aux(Y, temp_aux, P, equilibrium[1],
                                             v2, Xass2)

        lnK = fugl-fugv
        if it == (n-3):
            lnK3 = lnK
        elif it == (n-2):
            lnK2 = lnK
        elif it == (n-1):
            lnK1 = lnK
        elif it == n:
            it = 0
            itacc += 1
            dacc = gdem(lnK, lnK1, lnK2, lnK3)
            lnK += dacc
        K = np.exp(lnK)
        e1 = ((lnK-lnK_old)**2).sum()

    if e1 > K_tol and itacc == nacc and not singlephase:
        fobj = Gibbs_obj
        jac = True
        hess = None
        method = 'BFGS'
        y0 = beta*Y
        vsol = minimize(fobj, y0, args=(equilibrium, Z, temp_aux, P, model),
                        jac=jac, method=method, hess=hess, tol=K_tol)

        it2 += vsol.nit
        e1 = np.linalg.norm(vsol.jac)
        v = vsol.x
        l = Z - v
        beta = v.sum()
        v[v <= 1e-8] = 0
        l[l <= 1e-8] = 0
        Y = v / beta
        Y /= Y.sum()
        X = l/l.sum()

    if beta == 1.0:
        X = Y.copy()
    elif beta == 0.:
        Y = X.copy()

    if full_output:
        sol = {'T': T, 'P': P, 'beta': beta, 'error': e1, 'iter': it2,
               'X': X, 'v1': v1, 'Xass1': Xass1, 'state1': equilibrium[0],
               'Y': Y, 'v2': v2, 'Xass2': Xass2, 'state2': equilibrium[1]}
        out = EquilibriumResult(sol)
        return out

    return X, Y, beta



import numpy as np
from .monomer_aux import Xi, dXi_dxhi00,  d2Xi_dxhi00, dXi_dx


# Eq A20
def a2(xs, khs, xhixm, a2ij, epsij, f1, f2, f3):

    xi = Xi(xhixm, f1, f2, f3)
    a2 = a2ij * khs * (1 + xi) * epsij/2
    a = np.dot(xs, np.dot(a2, xs))

    return a


def da2_dxhi00(xs, khs, dkhs, xhixm, dxhim_dxhi00, da2ij, epsij, f1, f2, f3):

    xi, dxi = dXi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3)
    ctes = epsij / 2
    sum1, dsum1 = da2ij

    a2 = sum1 * khs * (1 + xi) * ctes
    da2 = sum1 * (1+xi) * dkhs + khs * (1 + xi) * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    da = np.array([a2, da2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a


def d2a2_dxhi00(xs, khs, dkhs, d2khs, xhixm, dxhim_dxhi00, d2a2ij, epsij,
                f1, f2, f3):

    xi, dxi, d2xi = d2Xi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3)

    ctes = epsij / 2
    sum1, dsum1, d2sum1 = d2a2ij

    a2 = sum1 * khs * (1 + xi) * ctes

    da2 = sum1 * (1+xi) * dkhs + khs * (1 + xi) * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    d2a2 = sum1 * (1 + xi) * d2khs
    d2a2 += sum1 * khs * d2xi
    d2a2 += d2sum1 * (1 + xi) * khs
    d2a2 += 2 * dkhs * ((1 + xi) * dsum1 + sum1 * dxi)
    d2a2 += 2 * khs * dsum1 * dxi
    d2a2 *= ctes

    da = np.array([a2, da2, d2a2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a


def da2_dx(xs, dxs_dx, khs, dkhs, xhixm, dxhim_dx, a2ij, da2ijx, epsij, f1,
           f2, f3):
    xi = Xi(xhixm, f1, f2, f3)
    dxi = dXi_dx(xhixm, dxhim_dx, f1, f2, f3)
    ctes = epsij/2

    a2 = a2ij * khs * (1 + xi) * ctes
    a = np.dot(xs, np.dot(a2, xs))

    da2 = np.multiply.outer(a2ij * (1+xi), dkhs).T
    da2 += a2ij * khs * dxi
    da2 += khs * (1 + xi) * da2ijx
    da2 *= ctes

    aux1 = xs * a2ij * khs * (1 + xi) * ctes
    suma1 = 2*np.sum(dxs_dx.T@aux1, axis=1)
    dax = np.matmul(np.matmul(da2, xs), xs)
    dax = xs@da2@xs + suma1

    return a, dax


def da2_dxxhi(xs, dxs_dx, khs, dkhs, dkhsx, xhixm, dxhim_dxhi00, dxhim_dx,
              da2ij, da2ijx, epsij, f1, f2, f3):

    ctes = epsij/2

    xi, dxi = dXi_dxhi00(xhixm, dxhim_dxhi00, f1, f2, f3)
    dxix = dXi_dx(xhixm, dxhim_dx, f1, f2, f3)

    sum1, dsum1 = da2ij

    a2 = sum1 * khs * (1 + xi) * ctes
    da2 = sum1 * (1+xi) * dkhs + khs * (1 + xi) * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    da = np.array([a2, da2])
    a = np.matmul(np.matmul(da, xs), xs)

    da2 = np.multiply.outer(sum1 * (1+xi), dkhsx).T
    da2 += sum1 * khs * dxix
    da2 += khs * (1 + xi) * da2ijx
    da2 *= ctes

    aux1 = xs * sum1 * khs * (1 + xi) * ctes
    suma1 = 2*np.sum(dxs_dx.T@aux1, axis=1)
    dax = np.matmul(np.matmul(da2, xs), xs)
    dax = xs@da2@xs + suma1

    return a, dax



import numpy as np


# Eq 47 to 50
def ki_chain(eta):
    eta2 = eta**2
    eta3 = eta**3
    eta4 = eta2**2
    eta_1 = (1 - eta)
    eta_13 = eta_1**3
    k0 = -np.log(eta_1) + (42*eta - 39*eta2 + 9*eta3 - 2*eta4)/(6*eta_13)
    k1 = (eta4 + 6*eta2 - 12*eta)/(2*eta_13)
    k2 = -3*eta2/(8*eta_1**2)
    k3 = (-eta4 + 3*eta2 + 3*eta)/(6*eta_13)
    return np.array([k0, k1, k2, k3])


def dki_chain(eta):
    eta2 = eta**2
    eta3 = eta**3
    eta4 = eta2**2
    eta_1 = (1 - eta)
    eta_13 = eta_1**3
    k0 = -np.log(eta_1) + (42*eta - 39*eta2 + 9*eta3 - 2*eta4)/(6*eta_13)
    k1 = (eta4 + 6*eta2 - 12*eta)/(2*eta_13)
    k2 = -3*eta2/(8*eta_1**2)
    k3 = (-eta4 + 3*eta2 + 3*eta)/(6*eta_13)

    dk0 = (24 + eta*(-6+eta*(3-7*eta+eta**2)))/(eta-1)**4/3
    dk1 = - (12 + eta*(2+eta)*(6-6*eta+eta2))/(eta-1)**4/2
    dk2 = 3*eta/(4*(-1+eta)**3)
    dk3 = (3+eta*(12+eta*(eta-3)*(eta-1)))/(6*(-1+eta)**4)
    return np.array([[k0, k1, k2, k3],
                     [dk0, dk1, dk2, dk3]])


def d2ki_chain(eta):
    eta2 = eta**2
    eta3 = eta**3
    eta4 = eta2**2
    eta_1 = (1 - eta)
    eta_13 = eta_1**3
    k0 = -np.log(eta_1) + (42*eta - 39*eta2 + 9*eta3 - 2*eta4)/(6*eta_13)
    k1 = (eta4 + 6*eta2 - 12*eta)/(2*eta_13)
    k2 = -3*eta2/(8*eta_1**2)
    k3 = (-eta4 + 3*eta2 + 3*eta)/(6*eta_13)

    dk0 = (24 + eta*(-6+eta*(3-7*eta+eta**2)))/(eta-1)**4/3
    dk1 = - (12 + eta*(2+eta)*(6-6*eta+eta2))/(eta-1)**4/2
    dk2 = 3*eta/(4*(-1+eta)**3)
    dk3 = (3+eta*(12+eta*(eta-3)*(eta-1)))/(6*(-1+eta)**4)

    d2k0 = (-30 + eta*(1+eta)*(4+eta))/(eta-1)**5
    d2k1 = 6*(5-2*eta*(eta-1))/(eta-1)**5
    d2k2 = -3*(1+2*eta)/(4*(eta-1)**4)
    d2k3 = (-4+eta*(eta-7))/((-1+eta)**5)
    return np.array([[k0, k1, k2, k3],
                     [dk0, dk1, dk2, dk3],
                     [d2k0, d2k1, d2k2, d2k3]])


# Eq 46
def gdHS(x0, eta):
    ks = ki_chain(eta)
    xs = np.array([1, x0, x0**2, x0**3])
    g = np.exp(np.dot(xs, ks))
    return g


def dgdHS_drho(x0, eta, deta_drho):
    dks = dki_chain(eta)
    xs = np.array([1., x0, x0**2, x0**3])
    dg = np.matmul(dks, xs)
    dg[0] = np.exp(dg[0])
    dg[1] *= dg[0]
    dg *= deta_drho[:2]
    return dg


def d2gdHS_drho(x0, eta, deta_drho):
    d2ks = d2ki_chain(eta)
    xs = np.array([1., x0, x0**2, x0**3])
    d2g = np.matmul(d2ks, xs)
    d2g[0] = np.exp(d2g[0])
    d2g[2] += d2g[1]**2
    d2g[2] *= d2g[0]
    d2g[1] *= d2g[0]
    d2g *= deta_drho[:3]
    return d2g


# Equation 64
def g1sigma(x0lambda, a1sb, da1m_deta,  eta, d, deta_drho, rho, eps, c, ms):

    ctes = 1/(2*np.pi*eps*ms*d**3)

    da1 = da1m_deta * deta_drho

    sum1 = np.dot(a1sb, x0lambda)
    g1 = ctes * (3.*da1 - c*sum1/rho)

    return g1


def dg1sigma_drho(x0lambda, da1sb, d2a1m_deta,  eta, d, drho, rho, eps, c, ms):
    ctes = 1/(2*np.pi*eps*ms*d**3)

    suma1 = np.matmul(da1sb, x0lambda)
    suma1 *= drho[:2]
    suma1 *= c

    d2a1m_drho = d2a1m_deta * drho
    dg1 = 3.*d2a1m_drho[1:] - suma1/rho
    dg1[1] += suma1[0]/rho**2
    dg1 *= ctes

    return dg1


def d2g1sigma_drho(x0lambda, d2a1sb, d3a1m_deta,  eta, d, drho, rho, eps, c,
                   ms):

    ctes = 1/(2*np.pi*eps*ms*d**3)

    suma1 = np.matmul(d2a1sb,   x0lambda)
    suma1 *= drho[:3]
    suma1 *= c

    d3a1m_drho = d3a1m_deta * drho
    d2g1 = 3*d3a1m_drho[1:] - suma1/rho
    d2g1[1] += suma1[0]/rho**2
    d2g1[2] += -2*suma1[0]/rho**3 + 2*suma1[1]/rho**2
    d2g1 *= ctes

    return d2g1


# Equation 65
def g2MCA(x0lambda, a1sb, da2m_new_deta, khs, eta,  d, deta_drho, rho, eps, c,
          ms):
    da2 = da2m_new_deta * deta_drho

    cte1 = (2*np.pi*eps**2*ms*d**3)
    cte2 = eps*c**2

    sum1 = np.dot(a1sb,  x0lambda)

    g2 = (3.*da2 - cte2 * khs * sum1 / rho)/cte1
    return g2


def dg2MCA_drho(x0lambda, a1sb, d2a2m_new_deta, dKhs, eta,  d, drho, rho, eps,
                c, ms):

    dKhs *= drho[:2]
    khs, dkhs = dKhs

    d2a2 = d2a2m_new_deta * drho[1:]

    cte1 = (2*np.pi*eps**2*ms*d**3)
    cte2 = eps*c**2

    suma1 = np.matmul(a1sb, x0lambda)
    suma1 *= drho[:2]
    suma1 *= cte2

    dg2 = 3*d2a2 - suma1 * khs/rho
    dg2[1] += suma1[0] * khs/rho**2 - suma1[0] * dkhs / rho
    dg2 /= cte1

    return dg2


def d2g2MCA_drho(x0lambda, a1sb, d3a2m_new_deta, d2Khs, eta,  d, drho, rho,
                 eps, c, ms):

    d2Khs *= drho[:3]
    khs, dkhs, d2khs = d2Khs

    d3a2 = d3a2m_new_deta * drho[1:]

    cte1 = (2*np.pi*eps**2*ms*d**3)
    cte2 = eps*c**2

    suma1 = np.matmul(a1sb, x0lambda)
    suma1 *= drho[:3]
    suma1 *= cte2
    sum1, dsum1, d2sum1 = suma1

    d2g2 = 3*d3a2 - suma1 * khs/rho
    d2g2[1] += sum1 * khs/rho**2 - sum1 * dkhs / rho
    d2g2[2] += -2*khs*sum1/rho**3 + 2*sum1*dkhs/rho**2 + 2*khs*dsum1/rho**2
    d2g2[2] += -2*dkhs*dsum1/rho - sum1*d2khs/rho
    d2g2 /= cte1

    return d2g2


phi7 = np.array([10., 10., 0.57, -6.7, -8])


# Equation 63
def gammac(x0, nsigma, alpha, tetha):
    gc = phi7[0]*(-np.tanh(phi7[1]*(phi7[2]-alpha))+1)
    gc *= nsigma*tetha*np.exp(phi7[3]*nsigma + phi7[4]*nsigma**2)
    return gc


def dgammac_deta(x03, nsigma, alpha, tetha):

    cte = phi7[0]*(-np.tanh(phi7[1]*(phi7[2]-alpha))+1)*tetha
    g = cte * np.exp(phi7[3]*nsigma + phi7[4]*nsigma**2)

    dg = np.full(2, g)
    dg[0] *= nsigma

    dg[1] *= (1. + nsigma*(phi7[3] + 2*phi7[4]*nsigma))
    dg[1] *= x03

    return dg


def d2gammac_deta(x03, nsigma, alpha, tetha):

    cte = phi7[0]*(-np.tanh(phi7[1]*(phi7[2]-alpha))+1)*tetha
    g = cte * np.exp(phi7[3]*nsigma+phi7[4]*nsigma**2)

    dg = np.full(3, g)
    dg[0] *= nsigma

    dg[1] *= (1. + nsigma*(phi7[3] + 2*phi7[4]*nsigma))
    dg[1] *= x03

    dg[2] *= (phi7[3]**2*nsigma+2*phi7[4]*nsigma*(3+2*phi7[4]*nsigma**2)+phi7[3]*(2.+4.*phi7[4]*nsigma**2))
    dg[2] *= x03**2

    return dg


def g2sigma(x03, nsigma, alpha, tetha, x0lambda, a1sb, da2m_new_deta, khs,
            eta, d, deta_drho, rho,  eps, c, ms):
    gc = gammac(x03, nsigma, alpha, tetha)
    g2 = g2MCA(x0lambda, a1sb, da2m_new_deta, khs, eta,  d, deta_drho, rho,
               eps, c, ms)
    return (1.+gc)*g2


def dg2sigma_drho(x03, nsigma, alpha, tetha, x0lambda, a1sb, d2a2m_new_deta,
                  dKhs, eta,  d, drho, rho,  eps, c, ms):

    gc_eval = dgammac_deta(x03, nsigma, alpha, tetha)
    gc_eval *= drho[:2]
    gc, dgc = gc_eval

    dg2mca = dg2MCA_drho(x0lambda, a1sb, d2a2m_new_deta, dKhs, eta, d, drho,
                         rho,  eps, c, ms)
    dg2 = dg2mca * (1. + gc)
    dg2[1] += dgc * dg2mca[0]

    return dg2


def d2g2sigma_drho(x03, nsigma, alpha, tetha, x0lambda, a1sb, d3a2m_new_deta,
                   d2Khs, eta,  d, drho, rho,  eps, c, ms):
    gc_eval = d2gammac_deta(x03, nsigma, alpha, tetha)
    gc_eval *= drho[:3]
    gc, dgc, d2gc = gc_eval

    d2g2mca = d2g2MCA_drho(x0lambda, a1sb, d3a2m_new_deta, d2Khs, eta, d,
                           drho, rho,  eps, c, ms)
    d2g2 = d2g2mca * (1. + gc)
    d2g2[1] += dgc * d2g2mca[0]
    d2g2[2] += d2gc*d2g2mca[0] + 2.*dgc*d2g2mca[1]

    return d2g2



import numpy as np


def association_config(eos):

    types = np.array(['B', 'P', 'N'])
    nozero = np.nonzero(eos.sites)
    types = types[nozero]
    ntypes = np.asarray(eos.sites)[nozero]
    nsites = len(types)
    S = np.array(eos.sites)
    S = S[S != 0]
    DIJ = np.zeros([nsites, nsites])
    int_i = []
    int_j = []
    for i in range(nsites):
        for j in range(nsites):
            bool1 = types[i] == 'B'
            bool2 = types[i] == 'P' and (types[j] == 'N' or types[j] == 'B')
            bool3 = types[i] == 'N' and (types[j] == 'P' or types[j] == 'B')
            if bool1 or bool2 or bool3:
                DIJ[i, j] = ntypes[j]
                int_i.append(i)
                int_j.append(j)

    indexabij = (int_i, int_j)
    diagasso = np.diag_indices(nsites)

    return S, DIJ, indexabij, nsites, diagasso


def Xass_solver(nsites, KIJ, diagasso, Xass0=None):

    if Xass0 is None:
        Xass = 0.2 * np.ones(nsites)
    else:
        Xass = 1. * Xass0
    omega = 0.2

    for i in range(5):
        fo = 1. / (1. + KIJ@Xass)
        dXass = (1 - omega) * (fo - Xass)
        Xass += dXass

    for i in range(10):
        KIJXass = KIJ@Xass
        dQ = (1./Xass - 1.) - KIJXass
        HIJ = - 1. * KIJ
        HIJ[diagasso] -= (1. + KIJXass)/Xass
        dXass = np.linalg.solve(HIJ, -dQ)
        Xass += dXass
        sucess = np.linalg.norm(dXass) < 1e-8
        if sucess:
            break

    return Xass


def Iab(di, rc, rd, eta,  sigma3):
    Kab = np.log((rc + 2*rd)/di)
    Kab *= 6*rc**3 + 18 * rc**2*rd - 24 * rd**3
    aux1 = (rc + 2 * rd - di)
    aux2 = (22*rd**2 - 5*rc*rd - 7*rd*di - 8*rc**2 + rc*di + di**2)
    Kab += aux1 * aux2
    Kab /= (72*rd**2 * sigma3)
    Kab *= 4 * np.pi * di**2

    gdhs = (1 - eta/2) / (1 - eta)**3

    Iab = Kab * gdhs

    return Iab


def dIab_drho(di, rc, rd, eta, deta_drho, sigma3):
    Kab = np.log((rc + 2*rd)/di)
    Kab *= 6*rc**3 + 18 * rc**2*rd - 24 * rd**3
    aux1 = (rc + 2 * rd - di)
    aux2 = (22*rd**2 - 5*rc*rd - 7*rd*di - 8*rc**2 + rc*di + di**2)
    Kab += aux1 * aux2
    Kab /= (72*rd**2 * sigma3)
    Kab *= 4 * np.pi * di**2

    gdhs = (1 - eta/2) / (1 - eta)**3
    dgdhs = (2.5 - eta) * deta_drho / (1 - eta)**4

    Iab = Kab * gdhs
    dIab = Kab * dgdhs
    return Iab, dIab


def d2Iab_drho(di, rc, rd, eta, deta_drho, sigma3):
    Kab = np.log((rc + 2*rd)/di)
    Kab *= 6*rc**3 + 18 * rc**2*rd - 24 * rd**3
    aux1 = (rc + 2 * rd - di)
    aux2 = (22*rd**2 - 5*rc*rd - 7*rd*di - 8*rc**2 + rc*di + di**2)
    Kab += aux1 * aux2
    Kab /= (72*rd**2 * sigma3)
    Kab *= 4 * np.pi * di**2

    gdhs = (1 - eta/2) / (1 - eta)**3
    dgdhs = (2.5 - eta) * deta_drho / (1 - eta)**4
    d2gdhs = 3 * (-3 + eta) * deta_drho**2 / (-1 + eta)**5

    Iab = Kab * gdhs
    dIab = Kab * dgdhs
    d2Iab = Kab * d2gdhs
    return Iab, dIab, d2Iab


def dXass_drho(rho, Xass, DIJ, Dabij, dDabij_drho, CIJ):
    brho = -(DIJ*(Dabij + rho * dDabij_drho))@Xass
    brho *= Xass**2
    dXass = np.linalg.solve(CIJ, brho)
    return dXass


def d2Xass_drho(rho, Xass, dXass_drho, DIJ, Dabij, dDabij_drho, d2Dabij_drho,
                CIJ):

    b2rho = np.sum(DIJ*(Xass*d2Dabij_drho + 2*dXass_drho*dDabij_drho), axis=1)
    b2rho *= - rho
    b2rho += 2 * (1/Xass - 1) / (rho**2)
    b2rho *= Xass**2
    b2rho += 2 * dXass_drho / (rho)
    b2rho += 2 * dXass_drho**2 / (Xass)

    d2Xass_drho = np.linalg.solve(CIJ, b2rho)
    return d2Xass_drho



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import minimize


def tpd(X, state, Z, T, P, model, v0=[None, None]):
    """
    Michelsen's Adimentional tangent plane function

    tpd (x, state, z, T, P, model)


    Parameters
    ----------
    X : array_like
        mole fraction array of trial fase
    state : string
        'L' for liquid phase, 'V' for vapour phase
    Z : array_like
        mole fraction array of overall mixture
    T :  float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        initial volume to solve fugacity, if supplied

    Returns
    -------
    tpd: float
        tpd distance

    """
    v1, v2 = v0
    logfugX, v1 = model.logfugef(X, T, P, state, v1)
    logfugZ, v2 = model.logfugef(Z, T, P, 'L', v2)
    di = np.log(Z) + logfugZ
    tpdi = X*(np.log(X) + logfugX - di)
    return np.sum(np.nan_to_num(tpdi))


def tpd_obj(a, temp_aux, P, di, model, state, vg):

    W = a**2/4.  # change from alpha to mole numbers
    w = W/W.sum()  # change to mole fraction
    # global vg
    logfugW, vg, _ = model.logfugef_aux(w, temp_aux, P, state, vg)

    dtpd = np.log(W) + logfugW - di
    tpdi = np.nan_to_num(W*(dtpd-1.))
    tpd = 1. + tpdi.sum()
    dtpd *= a/2
    return tpd, dtpd


def tpd_min(W, Z, T, P, model, stateW, stateZ, vw=None, vz=None):
    """

    Found a minimiun of Michelsen's Adimentional tangent plane function

    tpd_min (W, Z, T, P, model, stateW, stateZ)

    Parameters
    ----------
    W : array_like
        mole fraction array of trial fase
    Z : array_like
        mole fraction array of overall mixture
    T : float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object create from mixture, eos and mixrule
    stateW : string
        'L' for liquid phase, 'V' for vapor phase
    stateZ : string
        'L' for liquid phase, 'V' for vapor phase
    vw, vz: float, optional
        initial volume value to compute fugacities of phases

    Returns
    -------
    w : array_like
        molar fraction of minimum
    f : float
        minimized tpd distance

    """
    nc = model.nc
    if len(W) != nc or len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')
    # Fugacity of phase Zl

    temp_aux = model.temperature_aux(T)
    Z[Z < 1e-8] = 1e-8
    logfugZ, vz, Xassz = model.logfugef_aux(Z, temp_aux, P, stateZ, vz)
    di = np.log(Z) + logfugZ

    alpha0 = 2*W**0.5
    alpha0[alpha0 < 1e-8] = 1e-8  # To avoid negative compositions
    # global vg
    # vg = vw
    alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                     args=(temp_aux, P, di, model, stateW, vw))

    W = alpha.x**2/2
    w = W/W.sum()
    tpd = alpha.fun
    return w, tpd


def tpd_minimas(nmin, Z, T, P, model, stateW, stateZ, vw=None, vz=None):
    """

    Found nmin minimuns of Michelsen's Adimentional tangent plane function

    tpd_minimas (nmin, Z, T, P, model, stateW, stateZ)

    Parameters
    ----------
    nmin: int
        number of minimiuns to be founded
    Z : array_like
        mole fraction array of overall mixture
    T : float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object
        create from mixture, eos and mixrule
    stateW : string
        'L' for liquid phase, 'V' for vapour phase
    stateZ : string
        'L' for liquid phase, 'V' for vapour phase
    vw, vz : float, optional
        if supplied volume used as initial value to compute fugacities

    Returns
    -------
    all_minima: tuple
        molar fractions arrays of minimums
    f_minima: tuple
        minimized tpd distance

    """
    nc = model.nc
    if len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    temp_aux = model.temperature_aux(T)

    Z[Z < 1e-8] = 1e-8
    logfugZ, vz, Xassz = model.logfugef_aux(Z, temp_aux, P, stateZ, vz)
    di = np.log(Z) + logfugZ

    nc = model.nc
    all_minima = []
    f_minima = []

    # search from pures
    Id = np.eye(nc)
    alpha0 = 2*Id[0]**0.5
    alpha0[alpha0 < 1e-1] = 1e-1  # no negative or zero compositions
    # global vg
    vg = vw
    # Xassg = None
    alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                     args=(temp_aux, P, di, model, stateW, vg))
    W = alpha.x**2/4
    w = W/W.sum()  # normalized composition
    tpd = alpha.fun
    all_minima.append(w)
    f_minima.append(tpd)

    for i in range(1, nc):
        alpha0 = 2*Id[i]**0.5
        alpha0[alpha0 < 1e-1] = 1e-1
        vg = vw
        # Xassg = None
        alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                         args=(temp_aux, P, di, model, stateW, vg))
        W = alpha.x**2/4
        w = W/W.sum()  # normalized composition
        tpd = alpha.fun
        if alpha.success:
            add = np.any(np.all(np.isclose(all_minima, w, atol=1e-3), axis=1))
            if not add:
                f_minima.append(tpd)
                all_minima.append(w)
                if len(f_minima) == nmin:
                    return tuple(all_minima), np.array(f_minima)

    # random seach
    niter = 0
    while len(f_minima) < nmin and niter < (nmin+1):
        niter += 1
        Al = np.random.rand(nc)
        Al = Al/np.sum(Al)
        alpha0 = 2*Al**0.5
        alpha0[alpha0 < 1e-1] = 1e-1
        vg = vw
        alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                         args=(temp_aux, P, di, model, stateW, vg))
        W = alpha.x**2/4
        w = W/W.sum()  # normalized composition
        tpd = alpha.fun
        if alpha.success:
            add = np.any(np.all(np.isclose(all_minima, w, atol=1e-3), axis=1))
            if not add:
                f_minima.append(tpd)
                all_minima.append(w)
                if len(f_minima) == nmin:
                    return tuple(all_minima), np.array(f_minima)

    while len(f_minima) < nmin:
        all_minima.append(all_minima[0])
        f_minima.append(f_minima[0])

    return tuple(all_minima), np.array(f_minima)


def lle_init(Z, T, P, model, v0=None):
    """
    Minimize tpd function to initiate ELL at fixed T and P.

    Parameters
    ----------
    z : array_like
        overall molar fraction array
    T : float
        absolute temperature [K]
    P : float
        absolute pressure [Pa]
    model : object
        created from eos and mixture
    v0 : float, optional
        if supplied volume used as initial value to compute fugacities

    Returns
    -------
    x0s: tuple
        Contains two mol fractions arrays
    """
    x0s, tpd0 = tpd_minimas(2, Z, T, P, model, 'L', 'L', v0, v0)
    return x0s


def gmix(X, T, P, state, lnphi0, model, v0=None):
    lnphi, v = model.logfugmix(X, T, P, state, v0)
    gmix = lnphi
    gmix -= np.sum(X*lnphi0)
    gmix += np.sum(np.nan_to_num(X*np.log(X)))
    return gmix



from setuptools import setup, Extension
from Cython.Distutils import build_ext


cmdclass = {}
ext_modules = []


ext_modules += [Extension('SGTPy.coloc_cy',
                          ['SGTPy/src/coloc_cy.pyx']),
                Extension('SGTPy.sgt.cijmix_cy',
                          ['SGTPy/src/cijmix_cy.pyx'])]
cmdclass.update({'build_ext': build_ext})


setup(
  name='SGTPy',
  license='MIT',
  version='0.0.4',
  description='SAFT-VR-MIE EOS and SGT',
  author='Gustavo Chaparro Maldonado, Andres Mejia Matallana, Erich A. Muller',
  author_email='gustavochaparro@udec.cl',
  url='https://github.com/gustavochm/SGTPy',
  download_url='https://github.com/gustavochm/SGTPy.git',
  long_description=open('long_description.rst').read(),
  packages=['SGTPy', 'SGTPy.mixtures',  'SGTPy.pure',   'SGTPy.sgt',
            'SGTPy.equilibrium', 'SGTPy.fit'],
  cmdclass=cmdclass,
  ext_modules=ext_modules,
  install_requires=['numpy', 'scipy', 'cython'],
  platforms=["Windows", "Linux", "Mac OS", "Unix"],
  keywords=['SAFT-VR-Mie', 'SGT'],
  zip_safe=False
)



import numpy as np
from .monomer_aux import Xi, dXi, d2Xi
from .monomer_aux import dkHS, d2kHS, d3kHS

from .a1sB_monomer import da1B_eval, d2a1B_eval, d3a1B_eval
from .a1sB_monomer import x0lambda_eval

from .pertubaciones_eval import ahs, dahs_deta, d2ahs_deta
from .pertubaciones_eval import a1m
from .pertubaciones_eval import a2m, da2m_deta, d2a2m_deta
from .pertubaciones_eval import da2m_new_deta, d2a2m_new_deta, d3a2m_new_deta
from .pertubaciones_eval import a3m, da3m_deta, d2a3m_deta

from .monomer import amono, damono_drho, d2amono_drho
from .chain import achain, dachain_drho, d2achain_drho

from .association_aux import Xass_solver, Iab, dIab_drho
from .association_aux import d2Iab_drho, dXass_drho, d2Xass_drho

from .polarGV import Apol, dApol_drho, d2Apol_drho


def ares(self, rho, temp_aux, Xass0=None):
    beta, dia, tetha, x0, x03 = temp_aux[0:5]

    eta, deta = self.eta_bh(rho, dia)
    nsigma = self.eta_sigma(rho)

    # Parameters needed for evaluating the helmothlz contributions
    x0_a1, x0_a2, x0_a12, x0_a22 = x0lambda_eval(x0, self.lambda_a,
                                                 self.lambda_r, self.lambda_ar)
    a1sb_a1, a1sb_a2 = da1B_eval(x0, eta, self.lambda_a, self.lambda_r,
                                 self.lambda_ar, self.cctes, self.eps)
    dkhs = dkHS(eta)
    xi = Xi(x03, nsigma, self.f1, self.f2, self.f3)
    da2m_new = da2m_new_deta(x0_a2, a1sb_a2,  dkhs, self.c, self.eps)

    # Monomer contribution
    ahs_eval = ahs(eta)
    a1m_eval = a1m(x0_a1, a1sb_a1, self.c)
    a2m_eval = a2m(x0_a2, a1sb_a2[0], dkhs[0], xi, self.c, self.eps)
    a3m_eval = a3m(x03, nsigma, self.eps, self.f4, self.f5, self.f6)
    a_mono = amono(ahs_eval, a1m_eval[0], a2m_eval, a3m_eval, beta,
                   self.ms)

    # Chain contribution
    a_chain = achain(x0, eta, x0_a12, a1sb_a1[0], a1m_eval[1], x03, nsigma,
                     self.alpha, tetha, x0_a22, a1sb_a2[0], da2m_new,
                     dkhs[0], dia, deta, rho, beta,  self.eps, self.c,
                     self.ms, self.ring)

    # Total Helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:

        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0

        iab = Iab(dia, self.rcij, self.rdij, eta,  self.sigma3)
        Fab = temp_aux[5]
        Dab = self.sigma3 * Fab * iab
        Dabij = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab
        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)
        a += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[6]
        apolar = Apol(rho, eta, epsa, self.anij, self.bnij,
                      self.cnijk, self.mupolad2, self.npol, self.sigma3)
        a += apolar

    return a, Xass


def dares_drho(self, rho, temp_aux, Xass0=None):

    beta, dia, tetha, x0, x03 = temp_aux[0:5]

    eta, deta = self.eta_bh(rho, dia)
    nsigma = self.eta_sigma(rho)

    drho = np.array([1., deta, deta**2])

    # Parameters needed for evaluating the helmothlz contributions
    x0_a1, x0_a2, x0_a12, x0_a22 = x0lambda_eval(x0, self.lambda_a,
                                                 self.lambda_r,
                                                 self.lambda_ar)
    a1sb_a1, a1sb_a2 = d2a1B_eval(x0, eta, self.lambda_a, self.lambda_r,
                                  self.lambda_ar, self.cctes, self.eps)
    dkhs = d2kHS(eta)
    dxi = dXi(x03, nsigma, self.f1, self.f2, self.f3)
    da2m_new = d2a2m_new_deta(x0_a2, a1sb_a2, dkhs, self.c, self.eps)

    # Monomer contribution
    ahs_eval = dahs_deta(eta)
    a1m_eval = a1m(x0_a1, a1sb_a1, self.c)
    a2m_eval = da2m_deta(x0_a2, a1sb_a2[:2], dkhs[:2], dxi, self.c,
                         self.eps)
    a3m_eval = da3m_deta(x03, nsigma, self.eps, self.f4, self.f5, self.f6)
    a_mono = damono_drho(ahs_eval, a1m_eval[:2], a2m_eval, a3m_eval, beta,
                         drho[:2], self.ms)

    # Chain contribution
    a_chain = dachain_drho(x0, eta, x0_a12, a1sb_a1[:2], a1m_eval, x03,
                           nsigma, self.alpha, tetha, x0_a22, a1sb_a2[:2],
                           da2m_new, dkhs[:2], dia, drho, rho, beta,
                           self.eps, self.c, self.ms, self.ring)

    # Total helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        iab, diab = dIab_drho(dia, self.rcij, self.rdij, eta, deta,
                              self.sigma3)
        # Fab = np.exp(beta * self.eABij) - 1
        Fab = temp_aux[5]
        Dab = self.sigma3 * Fab * iab
        dDab = self.sigma3 * Fab * diab
        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])

        Dabij[self.indexabij] = Dab
        dDabij_drho[self.indexabij] = dDab
        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)

        CIJ = rho * np.tile(Xass**2, (self.nsites, 1)).T * Dabij * self.DIJ
        CIJ[self.diagasso] += 1.
        dXass = dXass_drho(rho, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)
        a[0] += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
        a[1] += np.dot(self.S, (1/Xass - 1/2) * dXass)
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[6]
        dapolar = dApol_drho(rho, eta, deta, epsa, self.anij,
                             self.bnij, self.cnijk, self.mupolad2,
                             self.npol, self.sigma3)
        a += dapolar

    return a, Xass


def d2ares_drho(self, rho, temp_aux, Xass0=None):

    beta, dia, tetha, x0, x03 = temp_aux[0:5]

    eta, deta = self.eta_bh(rho, dia)
    nsigma = self.eta_sigma(rho)

    drho = np.array([1., deta, deta**2, deta**3])

    # Parameters needed for evaluating the helmothlz contributions
    x0_a1, x0_a2, x0_a12, x0_a22 = x0lambda_eval(x0, self.lambda_a,
                                                 self.lambda_r,
                                                 self.lambda_ar)
    a1sb_a1, a1sb_a2 = d3a1B_eval(x0, eta, self.lambda_a, self.lambda_r,
                                  self.lambda_ar, self.cctes, self.eps)

    dkhs = d3kHS(eta)
    dxi = d2Xi(x03, nsigma, self.f1, self.f2, self.f3)
    da2m_new = d3a2m_new_deta(x0_a2, a1sb_a2,  dkhs, self.c, self.eps)

    # Monomer contribution
    ahs_eval = d2ahs_deta(eta)
    a1m_eval = a1m(x0_a1, a1sb_a1, self.c)
    a2m_eval = d2a2m_deta(x0_a2, a1sb_a2[:3], dkhs[:3], dxi, self.c,
                          self.eps)
    a3m_eval = d2a3m_deta(x03, nsigma, self.eps, self.f4, self.f5, self.f6)
    a_mono = d2amono_drho(ahs_eval, a1m_eval[:3], a2m_eval, a3m_eval, beta,
                          drho[:3], self.ms)

    # Chain contribution
    a_chain = d2achain_drho(x0, eta, x0_a12, a1sb_a1[:3], a1m_eval, x03,
                            nsigma, self.alpha, tetha, x0_a22, a1sb_a2[:3],
                            da2m_new, dkhs[:3], dia, drho, rho, beta,
                            self.eps, self.c, self.ms, self.ring)

    # Total Helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        iab, diab, d2iab = d2Iab_drho(dia, self.rcij, self.rdij, eta, deta,
                                      self.sigma3)

        # Fab = np.exp(beta * self.eABij) - 1.
        Fab = temp_aux[5]
        Dab = self.sigma3 * Fab * iab
        dDab = self.sigma3 * Fab * diab
        d2Dab = self.sigma3 * Fab * d2iab

        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])
        d2Dabij_drho = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab
        dDabij_drho[self.indexabij] = dDab
        d2Dabij_drho[self.indexabij] = d2Dab

        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)

        CIJ = rho * np.tile(Xass**2, (self.nsites, 1)).T * Dabij * self.DIJ
        CIJ[self.diagasso] += 1.
        dXass = dXass_drho(rho, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)
        d2Xass = d2Xass_drho(rho, Xass, dXass, self.DIJ, Dabij, dDabij_drho,
                             d2Dabij_drho, CIJ)
        a[0] += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
        a[1] += np.dot(self.S, (1/Xass - 1/2) * dXass)
        a[2] += np.dot(self.S, - (dXass/Xass)**2 + d2Xass * (1/Xass - 1/2))
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[6]
        dapolar = d2Apol_drho(rho, eta, deta, epsa, self.anij,
                              self.bnij, self.cnijk, self.mupolad2,
                              self.npol, self.sigma3)

        a += dapolar

    return a, Xass



from __future__ import division, print_function, absolute_import
from .mixtures import *
from .pure import *
from .saft import saftvrmie
from .mixture import component, mixture
from . import equilibrium
#from . import fit
from . import sgt
from .math import *
